"""Test pycontrails.physics.units module."""

from __future__ import annotations

from inspect import getmembers, isfunction

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails.physics import constants, units


@pytest.fixture(scope="module")
def rng():
    """Get random number generator."""
    return np.random.default_rng(12345)


def test_ft_m(rng):
    """Check `ft_to_m` and `m_to_ft` are bijective."""
    ft1 = rng.uniform(0, 10000, 10000)

    m1 = units.ft_to_m(ft1)
    ft2 = units.m_to_ft(m1)
    np.testing.assert_array_almost_equal(ft1, ft2)

    m2 = units.ft_to_m(ft2)
    np.testing.assert_array_almost_equal(m1, m2)


def test_knots_mps(rng):
    """Check that `m_per_s_to_knots` and `knots_to_m_per_s` are bijective."""
    mps1 = rng.uniform(0, 10000, 10000)

    knots1 = units.m_per_s_to_knots(mps1)
    mps2 = units.knots_to_m_per_s(knots1)
    np.testing.assert_array_almost_equal(mps1, mps2)

    knots2 = units.m_per_s_to_knots(mps2)
    np.testing.assert_array_almost_equal(knots1, knots2)


def test_rad_deg():
    """Check `degrees_to_radians` conversion on a few values."""
    d = 0
    r = units.degrees_to_radians(d)
    assert r == 0

    d = 180
    r = units.degrees_to_radians(d)
    assert r == pytest.approx(3.14159265)


def test_rad_deg_agree_np(rng):
    """Check that pycontrails implementation of agrees with numpy implementation."""
    x = rng.uniform(0, 1000, 10000)

    rad = units.radians_to_degrees(x)
    np_rad = np.rad2deg(x)
    np.testing.assert_array_equal(rad, np_rad)

    deg = units.degrees_to_radians(x)
    np_deg = np.deg2rad(x)
    np.testing.assert_array_equal(deg, np_deg)


def test_m_pl_bijective():
    """Check that the functions `pl_to_m` and `m_to_pl` are bijective."""
    m1 = np.arange(0, 50000)
    m2 = units.pl_to_m(units.m_to_pl(m1))
    np.testing.assert_allclose(m1, m2, atol=1e-11)

    pl1 = np.arange(1, 2000)
    pl2 = units.m_to_pl(units.pl_to_m(pl1))
    np.testing.assert_allclose(pl1, pl2, atol=1e-11)


def classical_pl_to_m(pl):
    """Classical implementation of pl_to_m."""
    pl_pa = pl * 100
    return (constants.T_msl / 0.0065) * (1 - (pl_pa / constants.p_surface) ** (1 / 5.255))


def test_pl_to_m_close_to_classical(rng):
    """Check that the classical conversion agrees for low altitudes."""
    p = rng.uniform(200, 1000, 10000)
    m1 = classical_pl_to_m(p)
    m2 = units.pl_to_m(p)
    np.testing.assert_allclose(m1, m2, rtol=1e-3)


def test_m_to_pl_int_and_array():
    """Check vectorized call agrees with naive loop.

    Calling two functions that use `np.piecewise` pattern.
    """
    arr = np.arange(15000)
    y1 = units.m_to_pl(arr)
    y2 = [units.m_to_pl(x) for x in arr]
    np.testing.assert_array_equal(y1, y2)

    arr = np.arange(100, 1000)
    y1 = units.pl_to_m(arr)
    y2 = [units.pl_to_m(x) for x in arr]
    np.testing.assert_array_equal(y1, y2)


def test_mach_tas(rng: np.random.Generator):
    """Check that the functions `tas_to_mach_number` and `mach_number_to_tas` are bijective."""
    T = rng.uniform(200, 300, 10000)
    tas1 = rng.uniform(200, 300, 10000)
    ma1 = units.tas_to_mach_number(tas1, T)

    # Ensure somewhat realistic values
    assert np.all(np.isfinite(ma1))
    assert np.all(ma1 < 1.1)
    assert np.all(ma1 > 0.5)
    assert ma1.mean() == pytest.approx(0.79, abs=0.01)

    tas2 = units.mach_number_to_tas(ma1, T)
    np.testing.assert_array_almost_equal(tas1, tas2, decimal=10)

    ma2 = units.tas_to_mach_number(tas2, T)
    assert np.all(np.isfinite(ma2))
    np.testing.assert_array_almost_equal(ma1, ma2, decimal=10)


@pytest.mark.parametrize(
    "func",
    [
        func
        for name, func in getmembers(units, isfunction)
        if name
        not in [
            "support_arraylike",
            "longitude_distance_to_m",
            "m_to_longitude_distance",
            "tas_to_mach_number",
            "mach_number_to_tas",
        ]
        and not name.startswith("_")
    ],
)
def test_arraylike_support(func):
    """Check that `unit` module functions support ArrayLike parameters."""
    x = 123
    y = 123.456789
    assert isinstance(func(x), float)
    assert isinstance(func(y), float)

    z = np.array([x, y])
    assert isinstance(func(z), np.ndarray)

    # Some functions can handle Series or DataArrays
    # Others convert to numpy
    da = xr.DataArray(z)
    assert isinstance(func(da), (xr.DataArray, np.ndarray))

    # Calling xr.apply_ufunc gives same result
    xr.testing.assert_equal(func(da), xr.apply_ufunc(func, da))

    s = pd.Series(z)
    assert isinstance(func(s), (pd.Series, np.ndarray))


@pytest.mark.parametrize("func", [units.pl_to_m, units.m_to_pl])
def test_handle_nan(func):
    """Check that `unit` module functions using `np.piecewise` pass NaN values through."""
    x = np.array([100, 200, np.nan, 300], dtype=np.float64)
    y = func(x)
    assert isinstance(y, np.ndarray)
    np.testing.assert_array_equal(np.isfinite(y), [True, True, False, True])
    assert np.isnan(x[2])
    assert np.isnan(y[2])



"""Non-volatile particulate matter (nvPM) calculations."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from pycontrails.physics import constants, jet, units
from pycontrails.utils.types import ArrayScalarLike

# ---------------------------------------------------------
# nvPM Mass Emissions Index: Formation and Oxidation Method
# ---------------------------------------------------------


def mass_emissions_index_fox(
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    fuel_flow_per_engine: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    pressure_ratio: float,
    *,
    comp_efficiency: float = 0.9,
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the black carbon mass emissions index using the Formation and Oxidation Method (FOX).

    Parameters
    ----------
    air_pressure: npt.NDArray[np.float_]
        Pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature: npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]
    true_airspeed: npt.NDArray[np.float_]
        True airspeed for each waypoint, [:math:`m s^{-1}`]
    fuel_flow_per_engine: npt.NDArray[np.float_]
        Fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, which is the fuel mass flow rate divided by
        the maximum fuel mass flow rate
    pressure_ratio: float
        Engine pressure ratio from the ICAO EDB
    comp_efficiency: float
        Engine compressor efficiency, assumed to be 0.9

    Returns
    -------
    npt.NDArray[np.float_]
        Black carbon mass emissions index, [:math:`mg \ kg_{fuel}^{-1}`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    """
    # Reference conditions: 100% thrust setting
    p_3_ref = jet.combustor_inlet_pressure(pressure_ratio, constants.p_surface, 1.0)
    t_3_ref = jet.combustor_inlet_temperature(
        comp_efficiency, constants.T_msl, constants.p_surface, p_3_ref
    )
    t_fl_ref = flame_temperature(t_3_ref)
    afr_ref = jet.air_to_fuel_ratio(1.0, cruise=False)
    fuel_flow_max = fuel_flow_per_engine / thrust_setting
    c_bc_ref = bc_mass_concentration_fox(fuel_flow_max, t_fl_ref, afr_ref)

    # Cruise conditions
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)
    t_2_cru = jet.compressor_inlet_temperature(air_temperature, mach_num)
    p_2_cru = jet.compressor_inlet_pressure(air_pressure, mach_num)
    p_3_cru = jet.combustor_inlet_pressure(pressure_ratio, p_2_cru, thrust_setting)
    t_3_cru = jet.combustor_inlet_temperature(comp_efficiency, t_2_cru, p_2_cru, p_3_cru)
    t_fl_cru = flame_temperature(t_3_cru)
    afr_cru = jet.air_to_fuel_ratio(thrust_setting, cruise=True, T_compressor_inlet=t_2_cru)
    c_bc_cru = bc_mass_concentration_cruise_fox(
        c_bc_ref=c_bc_ref,
        t_fl_cru=t_fl_cru,
        t_fl_ref=t_fl_ref,
        p_3_cru=p_3_cru,
        p_3_ref=p_3_ref,
        afr_cru=afr_cru,
        afr_ref=afr_ref,
    )
    q_exhaust_cru = exhaust_gas_volume_per_kg_fuel(afr_cru)
    return bc_mass_emissions_index(c_bc_cru, q_exhaust_cru)


def flame_temperature(t_3: ArrayScalarLike) -> ArrayScalarLike:
    """
    Calculate the flame temperature at the combustion chamber (t_fl).

    Parameters
    ----------
    t_3: ArrayScalarLike
        Combustor inlet temperature, [:math:`K`]

    Returns
    -------
    ArrayScalarLike
        Flame temperature at the combustion chamber, [:math:`K`]
    """
    return 0.9 * t_3 + 2120


def bc_mass_concentration_fox(
    fuel_flow: npt.NDArray[np.float_],
    t_fl: npt.NDArray[np.float_] | float,
    afr: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_]:
    """Calculate the black carbon mass concentration for ground conditions (``c_bc_ref``).

    This quantity is computed at the instrument sampling point without correcting
    for particle line losses.

    Parameters
    ----------
    fuel_flow: npt.NDArray[np.float_]
        Fuel mass flow rate, [:math:`kg s^{-1}`]
    t_fl: npt.NDArray[np.float_] | float
        Flame temperature at the combustion chamber, [:math:`K`]
    afr: npt.NDArray[np.float_] | float
        Air-to-fuel ratio

    Returns
    -------
    npt.NDArray[np.float_]:
        Black carbon mass concentration for ground conditions, [:math:`mg m^{-3}`]
    """
    return fuel_flow * (356 * np.exp(-6390 / t_fl) - 608 * afr * np.exp(-19778 / t_fl))


def bc_mass_concentration_cruise_fox(
    c_bc_ref: npt.NDArray[np.float_],
    t_fl_cru: npt.NDArray[np.float_],
    t_fl_ref: npt.NDArray[np.float_] | float,
    p_3_cru: npt.NDArray[np.float_],
    p_3_ref: npt.NDArray[np.float_] | float,
    afr_cru: npt.NDArray[np.float_],
    afr_ref: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_]:
    """Calculate the black carbon mass concentration for cruise conditions (``c_bc_cru``).

    This quantity is computed at the instrument sampling point without correcting
    for particle line losses.

    Parameters
    ----------
    c_bc_ref: npt.NDArray[np.float_]
        Black carbon mass concentration at reference conditions, [:math:`mg m^{-3}`]
    t_fl_cru: npt.NDArray[np.float_]
        Flame temperature at cruise conditions, [:math:`K`]
    t_fl_ref: npt.NDArray[np.float_] | float
        Flame temperature at reference conditions, [:math:`K`]
    p_3_cru: npt.NDArray[np.float_]
        Combustor inlet pressure at cruise conditions, [:math:`Pa`]
    p_3_ref: npt.NDArray[np.float_] | float
        Combustor inlet pressure at reference conditions, [:math:`Pa`]
    afr_cru: npt.NDArray[np.float_]
        Air-to-fuel ratio at cruise conditions
    afr_ref: npt.NDArray[np.float_] | float
        Air-to-fuel ratio at reference conditions

    Returns
    -------
    npt.NDArray[np.float_]:
        Black carbon mass concentration for cruise conditions, [:math:`mg m^{-3}`]
    """
    scaling_factor = dopelheuer_lecht_scaling_factor(
        t_fl_cru=t_fl_cru,
        t_fl_ref=t_fl_ref,
        p_3_cru=p_3_cru,
        p_3_ref=p_3_ref,
        afr_cru=afr_cru,
        afr_ref=afr_ref,
    )
    return c_bc_ref * scaling_factor


def dopelheuer_lecht_scaling_factor(
    t_fl_cru: npt.NDArray[np.float_],
    t_fl_ref: npt.NDArray[np.float_] | float,
    p_3_cru: npt.NDArray[np.float_],
    p_3_ref: npt.NDArray[np.float_] | float,
    afr_cru: npt.NDArray[np.float_],
    afr_ref: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_]:
    """Estimate scaling factor to convert the reference BC mass concentration from ground to cruise.

    Parameters
    ----------
    t_fl_cru: npt.NDArray[np.float_]
        Flame temperature at cruise conditions, [:math:`K`]
    t_fl_ref: npt.NDArray[np.float_] | float
        Flame temperature at reference conditions, [:math:`K`]
    p_3_cru: npt.NDArray[np.float_]
        Combustor inlet pressure at cruise conditions, [:math:`Pa`]
    p_3_ref: npt.NDArray[np.float_] | float
        Combustor inlet pressure at reference conditions, [:math:`Pa`]
    afr_cru: npt.NDArray[np.float_]
        Air-to-fuel ratio at cruise conditions
    afr_ref: npt.NDArray[np.float_] | float
        Air-to-fuel ratio at reference conditions

    Returns
    -------
    npt.NDArray[np.float_]
        Dopelheuer & Lecht scaling factor

    References
    ----------
    - :cite:`dopelheuerInfluenceEnginePerformance1998`
    """
    exp_term = np.exp(20000 / t_fl_cru) / np.exp(20000 / t_fl_ref)
    return (afr_ref / afr_cru) ** 2.5 * (p_3_cru / p_3_ref) ** 1.35 * exp_term


# ----------------------------------------------------------------------------
# nvPM Mass Emissions Index: "Improved" Formation and Oxidation Method (ImFOX)
# ----------------------------------------------------------------------------


def mass_emissions_index_imfox(
    fuel_flow_per_engine: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    fuel_hydrogen: float,
) -> npt.NDArray[np.float_]:
    r"""Calculate the BC mass EI using the "Improved" Formation and Oxidation Method (ImFOX).

    Parameters
    ----------
    fuel_flow_per_engine: npt.NDArray[np.float_]
        Fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, which is the fuel mass flow rate divided by the
        maximum fuel mass flow rate
    fuel_hydrogen: float
        Percentage of hydrogen mass content in the fuel (13.8% for conventional Jet A-1 fuel)

    Returns
    -------
    npt.NDArray[np.float_]
        Black carbon mass emissions index, [:math:`mg \ kg_{fuel}^{-1}`]

    References
    ----------
    - :cite:`abrahamsonPredictiveModelDevelopment2016`
    """
    afr_cru = air_to_fuel_ratio_imfox(thrust_setting)
    t_4_cru = turbine_inlet_temperature_imfox(afr_cru)
    c_bc_cru = bc_mass_concentration_imfox(
        fuel_flow_per_engine, afr_cru, t_4_cru, fuel_hydrogen=fuel_hydrogen
    )
    q_exhaust_cru = exhaust_gas_volume_per_kg_fuel(afr_cru)
    return bc_mass_emissions_index(c_bc_cru, q_exhaust_cru)


def air_to_fuel_ratio_imfox(thrust_setting: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """Calculate the air-to-fuel ratio at cruise conditions via Abrahamson's method.

    See Eq. (11) in :cite:`abrahamsonPredictiveModelDevelopment2016`.

    Parameters
    ----------
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, which is the fuel mass flow rate divided by
        the maximum fuel mass flow rate

    Returns
    -------
    npt.NDArray[np.float_]
        Air-to-fuel ratio at cruise conditions

    References
    ----------
    - :cite:`abrahamsonPredictiveModelDevelopment2016`
    """
    return 55.4 - 30.8 * thrust_setting


def turbine_inlet_temperature_imfox(afr: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """Calculate the turbine inlet temperature using Abrahamson's method.

    See Eq. (13) in :cite:`abrahamsonPredictiveModelDevelopment2016`.

    Parameters
    ----------
    afr: npt.NDArray[np.float_]
        air-to-fuel ratio at cruise conditions

    Returns
    -------
    npt.NDArray[np.float_]
        turbine inlet temperature, [:math:`K`]

    References
    ----------
    - :cite:`abrahamsonPredictiveModelDevelopment2016`
    """
    return 490 + 42266 / afr


def bc_mass_concentration_imfox(
    fuel_flow_per_engine: npt.NDArray[np.float_],
    afr: npt.NDArray[np.float_],
    t_4: npt.NDArray[np.float_],
    fuel_hydrogen: float,
) -> npt.NDArray[np.float_]:
    """Calculate the BC mass concentration for ground and cruise conditions with ImFOX methodology.

    This quantity is computed at the instrument sampling point without
    correcting for particle line losses.

    Parameters
    ----------
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    afr: npt.NDArray[np.float_]
        air-to-fuel ratio
    t_4: npt.NDArray[np.float_]
        turbine inlet temperature, [:math:`K`]
    fuel_hydrogen: float
        percentage of hydrogen mass content in the fuel (13.8% for conventional Jet A-1 fuel)

    Returns
    -------
    npt.NDArray[np.float_]
        Black carbon mass concentration, [:math:`mg m^{-3}`]
    """
    exp_term = np.exp(13.6 - fuel_hydrogen)
    formation_term = 295 * np.exp(-6390 / t_4)
    oxidation_term = 608 * afr * np.exp(-19778 / t_4)
    return fuel_flow_per_engine * exp_term * (formation_term - oxidation_term)


# ---------------------------------------------------------
# Commonly Used Black Carbon Mass Emissions Index Functions
# ---------------------------------------------------------


def exhaust_gas_volume_per_kg_fuel(afr: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """
    Calculate the volume of exhaust gas per mass of fuel burnt.

    Parameters
    ----------
    afr: npt.NDArray[np.float_]
        Air-to-fuel ratio

    Returns
    -------
    npt.NDArray[np.float_]
        Volume of exhaust gas per mass of fuel burnt, [:math:`m^{3}/kg_{fuel}`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    """
    return 0.776 * afr + 0.877


def bc_mass_emissions_index(
    c_bc: npt.NDArray[np.float_], q_exhaust: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the black carbon mass emissions index.

    Parameters
    ----------
    c_bc: npt.NDArray[np.float_]
        Black carbon mass concentration, [:math:`mg m^{-3}`]
    q_exhaust: npt.NDArray[np.float_]
        Volume of exhaust gas per mass of fuel burnt, [:math:`m^{3}/kg_{fuel}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Black carbon mass emissions index, [:math:`mg/kg_{fuel}`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    """
    return c_bc * q_exhaust


# ----------------------------------------------------------
# nvPM Number Emissions Index: Fractal Aggregates (FA) model
# ----------------------------------------------------------


def geometric_mean_diameter_sac(
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    pressure_ratio: float,
    q_fuel: float,
    *,
    comp_efficiency: float = 0.9,
    delta_loss: float = 5.75,
    cruise: bool = True,
) -> npt.NDArray[np.float_]:
    r"""Calculate the BC GMD for singular annular combustor (SAC) engines.

    The BC (black carbon) GMD (geometric mean diameter) is estimated using
    the non-dimensionalized engine thrust setting, the ratio of turbine inlet
    to the compressor inlet temperature (``t4_t2``).

    Parameters
    ----------
    air_pressure: npt.NDArray[np.float_]
        Pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature: npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]
    true_airspeed: npt.NDArray[np.float_]
        True airspeed for each waypoint, [:math:`m s^{-1}`]
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, which is the fuel mass flow rate divided by the
        maximum fuel mass flow rate
    pressure_ratio: float
        Engine pressure ratio from the ICAO EDB
    q_fuel : float
        Lower calorific value (LCV) of fuel, [:math:`J \ kg_{fuel}^{-1}`]
    comp_efficiency: float
        Engine compressor efficiency (assumed to be 0.9)
    delta_loss: float
        Correction factor accounting for particle line losses (assumed to be 5.75 nm), [:math:`nm`]
    cruise: bool
        Set to true when the aircraft is not on the ground.

    Returns
    -------
    npt.NDArray[np.float_]
        black carbon geometric mean diameter, [:math:`nm`]

    References
    ----------
    - :cite:`teohMitigatingClimateForcing2020`
    """
    t4_t2 = jet.thrust_setting_nd(
        true_airspeed,
        thrust_setting,
        air_temperature,
        air_pressure,
        pressure_ratio,
        q_fuel,
        comp_efficiency=comp_efficiency,
        cruise=cruise,
    )
    return (2.5883 * t4_t2**2) - (5.3723 * t4_t2) + 16.721 - delta_loss


def number_emissions_index_fractal_aggregates(
    nvpm_ei_m: npt.NDArray[np.float_],
    gmd: npt.NDArray[np.float_],
    *,
    gsd: float | npt.NDArray[np.float_] = 1.80,
    rho_bc: float = 1770,
    k_tem: float = 1.621e-5,
    d_tem: float = 0.39,
    d_fm: float = 2.76,
) -> npt.NDArray[np.float_]:
    """
    Estimate the black carbon number emission index using the fractal aggregates (FA) model.

    The FA model estimates the number emissions index from the mass emissions index,
    particle size distribution, and morphology.

    Parameters
    ----------
    nvpm_ei_m: npt.NDArray[np.float_]
        Black carbon mass emissions index, [:math:`kg/kg_{fuel}`]
    gmd: npt.NDArray[np.float_]
        Black carbon geometric mean diameter, [:math:`m`]
    gsd: float
        Black carbon geometric standard deviation (assumed to be 1.80)
    rho_bc: float
        Black carbon material density (1770 kg/m**3), [:math:`kg m^{-3}`]
    k_tem: float
        Transmission electron microscopy prefactor coefficient (assumed to be 1.621e-5)
    d_tem: float
        Transmission electron microscopy exponent coefficient (assumed to be 0.39)
    d_fm: float
        Mass-mobility exponent (assumed to be 2.76)

    Returns
    -------
    npt.NDArray[np.float_]
        Black carbon number emissions index, [:math:`kg_{fuel}^{-1}`]

    References
    ----------
    - FA model: :cite:`teohMethodologyRelateBlack2019`
    - ``gmd``, ``gsd``, ``d_fm``: :cite:`teohMitigatingClimateForcing2020`
    - ``rho_bc``: :cite:`parkMeasurementInherentMaterial2004`
    - ``k_tem``, ``d_tem``: :cite:`dastanpourObservationsCorrelationPrimary2014`
    """
    phi = 3 * d_tem + (1 - d_tem) * d_fm
    exponential_term = np.exp(0.5 * phi**2 * np.log(gsd) ** 2)
    denom = rho_bc * (np.pi / 6) * k_tem ** (3 - d_fm) * gmd**phi * exponential_term
    return nvpm_ei_m / denom


# -------------------------------------------------------------
# Scale nvPM emissions indices due to sustainable aviation fuel
# -------------------------------------------------------------


def nvpm_number_ei_pct_reduction_due_to_saf(
    hydrogen_content: float | npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Adjust nvPM number emissions index to account for the effects of sustainable aviation fuels.

    Parameters
    ----------
    hydrogen_content: float
        The percentage of hydrogen mass content in the fuel.
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, where the equivalent fuel mass flow rate per engine at
        sea level, :math:`[0 - 1]`.

    Returns
    -------
    npt.NDArray[np.float_]
        Percentage reduction in nvPM number emissions index

    References
    ----------
    - :cite:`teohTargetedUseSustainable2022`
    - :cite:`bremEffectsFuelAromatic2015`
    """
    a0 = -114.21
    a1 = 1.06
    a2 = 0.5
    return _template_saf_reduction(hydrogen_content, thrust_setting, a0, a1, a2)


def nvpm_mass_ei_pct_reduction_due_to_saf(
    hydrogen_content: float | npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Adjust nvPM mass emissions index to account for the effects of sustainable aviation fuels.

    For fuel with hydrogen mass content > 14.3, the adjustment factor is adopted from
    Teoh et al. (2022), which was used to calculate the change in nvPM EIn.

    Parameters
    ----------
    hydrogen_content: float
        The percentage of hydrogen mass content in the fuel.
    thrust_setting: npt.NDArray[np.float_]
        Engine thrust setting, where the equivalent fuel mass flow rate per engine at
        sea level, :math:`[0 - 1]`.

    Returns
    -------
    npt.NDArray[np.float_]
        Percentage reduction in nvPM number emissions index

    References
    ----------
    - :cite:`teohTargetedUseSustainable2022`
    - :cite:`bremEffectsFuelAromatic2015`
    """
    a0 = -124.05
    a1 = 1.02
    a2 = 0.6
    return _template_saf_reduction(hydrogen_content, thrust_setting, a0, a1, a2)


def _template_saf_reduction(
    hydrogen_content: float | npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    a0: float,
    a1: float,
    a2: float,
) -> npt.NDArray[np.float_]:
    # Thrust setting cannot be computed when engine data is not provided in
    # the ICAO EDB, so set default to 45% thrust.
    thrust_setting = np.nan_to_num(thrust_setting, nan=0.45)
    delta_h = hydrogen_content - 13.80
    d_nvpm_ein_pct = (a0 + a1 * (thrust_setting * 100.0)) * delta_h

    # Adjust when delta_h is large
    if isinstance(delta_h, np.ndarray):
        filt = delta_h > 0.5
        d_nvpm_ein_pct[filt] *= np.exp(0.5 * (a2 - delta_h[filt]))
    elif delta_h > 0.5:
        d_nvpm_ein_pct *= np.exp(0.5 * (a2 - delta_h))

    d_nvpm_ein_pct.clip(min=-90.0, max=0.0, out=d_nvpm_ein_pct)
    return d_nvpm_ein_pct



"""Test `MetDatset` and `MetDataArray` data structures."""

from __future__ import annotations

import json
import pathlib
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import DiskCacheStore, MetDataArray, MetDataset, MetVariable
from pycontrails.core import cache as cache_module
from pycontrails.core.met import originates_from_ecmwf, shift_longitude
from pycontrails.datalib.ecmwf import ERA5
from tests import OPEN3D_AVAILABLE

DISK_CACHE_DIR = cache_module._get_user_cache_dir()


def test_sl_path_to_met(met_ecmwf_sl_path: str) -> None:
    """Confirm `MetDataset` constructor converts `Dataset` with correct conventions."""
    ds = xr.open_dataset(met_ecmwf_sl_path)
    assert set(ds.variables).issubset({"longitude", "latitude", "level", "time", "sp"})
    assert list(ds.data_vars) == ["sp"]

    mds = MetDataset(ds)
    assert mds.dim_order == ["longitude", "latitude", "level", "time"]

    # longitude and latitude correctly translated
    assert np.all(mds.data["longitude"].values >= -180)
    assert np.all(mds.data["longitude"].values < 180)
    assert np.all(mds.data["latitude"].values <= 90)
    assert np.all(mds.data["latitude"].values >= -90)

    # consistent shapes between coordinates and underlying np.array of values
    shapes: tuple[int, ...] = tuple()
    for coord in mds.dim_order:
        # every coordinate is ascending
        assert np.all(np.diff(mds.data[coord]).astype(float) > 0)
        shapes = shapes + mds.data[coord].shape

    # operators commute
    assert isinstance(mds["sp"], MetDataArray)
    assert isinstance(mds["sp"].data, xr.DataArray)
    assert isinstance(mds.data["sp"], xr.DataArray)
    xr.testing.assert_equal(mds["sp"].data, mds.data["sp"])

    assert shapes == mds["sp"].data.values.shape

    # wrap longitude
    mds = MetDataset(ds, wrap_longitude=True)

    # longitude and latitude correctly translated, but not wrapped
    assert np.all(mds.data["longitude"].values <= 200)
    assert np.all(mds.data["longitude"].values >= -185)

    assert np.all(mds.data["sp"][dict(longitude=0)] == mds.data["sp"][dict(longitude=-2)])
    assert np.all(mds.data["sp"][dict(longitude=1)] == mds.data["sp"][dict(longitude=-1)])

    assert not mds.is_zarr


def test_pl_path_to_met(met_ecmwf_pl_path: str) -> None:
    """Confirm `MetDataset` constructor converts `Dataset` with correct conventions."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    assert set(ds.variables).issubset(
        {"longitude", "latitude", "level", "time", "t", "r", "q", "ciwc"}
    )
    assert list(ds.data_vars) == ["t", "r", "q", "ciwc"]

    mds = MetDataset(ds)
    assert mds.dim_order == ["longitude", "latitude", "level", "time"]

    # longitude and latitude correctly translated
    assert np.all(mds.data["longitude"].values < 180)
    assert np.all(mds.data["longitude"].values >= -180)
    assert np.all(mds.data["latitude"].values <= 90)
    assert np.all(mds.data["latitude"].values >= -90)

    # consistent shapes between coordinates and underlying np.array of values
    shapes: tuple[int, ...] = tuple()
    for coord in mds.dim_order:
        # every coordinate is ascending
        assert np.all(np.diff(mds.data[coord]).astype(float) > 0)
        shapes = shapes + mds.data[coord].shape
    assert shapes == mds.data["t"].values.shape

    with pytest.raises(ValueError, match="Set 'copy=True' when using 'wrap_longitude=True'."):
        MetDataset(ds, wrap_longitude=True, copy=False)

    with pytest.raises(ValueError, match="Coordinate 'latitude' not sorted."):
        MetDataset(ds, copy=False)

    assert not mds.is_zarr


@pytest.mark.parametrize("variable", ["t", "q", "r", "sp"])
def test_metdataarray_constructor(
    met_ecmwf_sl_path: str, met_ecmwf_pl_path: str, variable: str
) -> None:
    """Check `MetDataArray` constructor applies conventions."""
    if variable == "sp":
        ds = xr.open_dataset(met_ecmwf_sl_path)
    else:
        ds = xr.open_dataset(met_ecmwf_pl_path)

    da = ds[variable]
    mda = MetDataArray(da, name="orcas")

    assert mda.name == "orcas"
    assert "MetDataArray" in repr(mda)
    assert isinstance(mda.hash, str)
    assert isinstance(mda.shape, tuple)
    assert not mda.is_wrapped
    assert mda.shape[0] == 15

    mda2 = MetDataArray(da, wrap_longitude=True)
    assert mda2.is_wrapped
    assert mda2.shape[0] == 17

    with pytest.raises(ValueError, match="Set 'copy=True' when using 'wrap_longitude=True'."):
        MetDataArray(mda.data, wrap_longitude=True, copy=False)

    # Cannot instantiate MetDataArray without time or level coord
    # This hacked DataArray only contains latitude and longitude
    insufficient_da = da["latitude"] + da["longitude"]
    with pytest.raises(ValueError, match="Meteorology data must contain dimension 'level'"):
        MetDataArray(insufficient_da)


@pytest.mark.parametrize("wrap_longitude", [True, False])
def test_pl_path_to_met_dataset(met_ecmwf_pl_path: str, wrap_longitude: bool):
    """Confirm that DataArray selection is consistent with attr MetDataArray.data."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds, wrap_longitude=wrap_longitude)

    # TODO: I have no idea why this `sel` doesn't work all at once. Likely an xarray bug
    sel = dict(latitude=15.0, longitude=25.0, level=225)
    time_sel = dict(time="2019-05-31 05:00:00")
    val2 = mds["t"].data.loc[sel].loc[time_sel].values
    val1 = ds["t"].loc[sel].loc[time_sel].values
    assert val1 == val2

    sel = dict(latitude=15.0, longitude=25.0, level=225)
    time_sel = dict(time="2019-05-31 05:00:00")
    val2 = mds["q"].data.loc[sel].loc[time_sel].values
    val1 = ds["q"].loc[sel].loc[time_sel].values
    np.testing.assert_approx_equal(val1, val2)

    assert "MetDataset" in repr(mds)

    # hash
    assert isinstance(mds.hash, str)


@pytest.mark.parametrize(
    "lon,lat,match",
    [
        (np.linspace(-200, 200, 10), np.linspace(-100, 100, 10), "Latitude values must"),
        (np.linspace(-400, 200, 10), np.linspace(-80, 80, 10), "Shift to WGS84"),
    ],
)
def test_preprocess_dims_error(lon: np.ndarray, lat: np.ndarray, match: str):
    """Confirm ValueErrors if coordinates are not WGS84."""
    bad_da = xr.DataArray(
        data=1,
        dims=["longitude", "latitude", "level", "time"],
        coords={
            "longitude": lon,
            "latitude": lat,
            "time": [datetime(2000, 1, 1, h) for h in range(5)],
            "level": np.arange(100, 500, 100),
        },
    )

    with pytest.raises(ValueError, match=match):
        MetDataArray(bad_da)


def test_wrap_longitude(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> None:
    # wrap datasets
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds, wrap_longitude=True)

    # longitude and latitude correctly translated, but not wrapped
    assert np.all(mds.data["longitude"].values <= 200)
    assert np.all(mds.data["longitude"].values >= -185)

    assert np.all(mds["t"].data[dict(longitude=0)] == mds["t"].data[dict(longitude=-2)])
    assert np.all(mds["t"].data[dict(longitude=1)] == mds["t"].data[dict(longitude=-1)])

    # wrap dataarrays
    mda = MetDataArray(zero_like_da, wrap_longitude=True)

    # longitude and latitude correctly translated, but not wrapped
    assert np.all(mda.data["longitude"].values <= 200)
    assert np.all(mda.data["longitude"].values >= -185)

    assert np.all(mda.data[dict(longitude=0)] == mda.data[dict(longitude=-2)])
    assert np.all(mda.data[dict(longitude=1)] == mda.data[dict(longitude=-1)])

    da = zero_like_da.copy()
    da2 = da[dict(longitude=-1)].assign_coords(longitude=180)
    da = xr.concat([da, da2], dim="longitude")

    # automatic wrapping inspection
    mda = MetDataArray(da)
    assert mda.is_wrapped


def test_shift_longitude(zero_like_da: xr.DataArray) -> None:
    da = zero_like_da.copy()

    lons = da["longitude"].values
    lons[lons < 0] = lons[lons < 0] + 360

    da = da.assign_coords(longitude=lons)
    assert np.all(da["longitude"].values >= 0)

    da2 = shift_longitude(da)

    # longitude correctly translated
    assert np.all(da2["longitude"].values <= 180)
    assert np.all(da2["longitude"].values >= -180)

    # longitude sorted
    assert np.all(np.diff(da2["longitude"].values) > 0)


@pytest.fixture
def zero_like_da() -> xr.DataArray:
    return xr.DataArray(
        data=0.0,
        dims=["longitude", "latitude", "level", "time"],
        coords={
            "longitude": np.arange(-180, 180, 1.0),
            "latitude": np.arange(-90, 90.1, 1.0),
            "time": [datetime(2000, 1, 1, h) for h in range(5)],
            "level": np.arange(100, 500, 100),
        },
    )


# a few arbitrary points with which to build sparse binary
@pytest.fixture(
    params=[
        {"lon": 13, "lat": 5, "level": 100},
        {"lon": 21, "lat": -8, "level": 200},
        {"lon": -34, "lat": 3, "level": 300},
        {"lon": -144, "lat": 42, "level": 400},
        {"lon": 27, "lat": 42, "level": 400},
        {"lon": 18, "lat": 0, "level": 400},
    ]
)
def sparse_binary(zero_like_da: xr.DataArray, request: Any) -> tuple[MetDataArray, dict]:
    """Return a MetDataArray with exactly one nonzero value at each time slice.

    Data has integer longitude and latitude values, 4 level coordinates, and 5 time coordinates.
    """
    # a single arbitrary positive point
    da = zero_like_da
    lon = request.param["lon"]
    lat = request.param["lat"]
    level = request.param["level"]
    point = (da["latitude"] == lat) & (da["longitude"] == lon) & (da["level"] == level)
    da = xr.where(point, 1.0, da)
    return MetDataArray(da), request.param


@pytest.fixture
def island_binary(zero_like_da: xr.DataArray) -> MetDataArray:
    # create 9 x 9 island

    da = zero_like_da
    island = (np.abs(da["latitude"]) < 5) & (np.abs(da["longitude"]) < 5)
    da = xr.where(island, 1.0, da)
    return MetDataArray(da)


@pytest.fixture
def antimeridian_binary(zero_like_da: xr.DataArray) -> MetDataArray:
    # create 9 x 9 island spanning antimeridian

    da = zero_like_da
    island = (np.abs(da["latitude"]) < 5) & ((da["longitude"] < -175) | (da["longitude"] > 175))
    da = xr.where(island, 1, da)
    return MetDataArray(da, wrap_longitude=True)


@pytest.fixture(params=["t", "r", "q"])
def median_binary(met_ecmwf_pl_path: str, request: Any) -> MetDataArray:
    # points with values above the median get 1, else 0
    ds = xr.open_dataset(met_ecmwf_pl_path)
    da = ds[request.param]
    da = (da > da.median()).astype(int)
    return MetDataArray(da)


def _generate_waypoints(
    x0: float, x1: float, y0: float, y1: float, z0: float, z1: float, t0: str, t1: str
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Generate waypoints spanning various diagonals of some 4d box."""
    n_points = 100

    x = np.linspace(x0, x1, n_points)
    y = np.linspace(y0, y1, n_points)
    z = np.linspace(z0, z1, n_points)
    t = pd.date_range(t0, t1, n_points)

    _x = np.concatenate([x, np.flip(x), x, x, x])
    _y = np.concatenate([y, y, np.flip(y), y, y])
    _z = np.concatenate([z, z, z, np.flip(z), z])
    _t = np.concatenate([t, t, t, t, np.flip(t)])

    return _x, _y, _z, _t


@pytest.fixture
def pl_path_to_doubled_time_mds(met_ecmwf_pl_path: str) -> MetDataset:
    return MetDataset(xr.open_dataset(met_ecmwf_pl_path))


@pytest.fixture
def sl_path_to_tripled_time_mds(met_ecmwf_sl_path: str) -> MetDataset:
    """Triple the number of time variables in the Dataset sitting at `met_ecmwf_sl_path`."""
    ds1 = xr.open_dataset(met_ecmwf_sl_path)

    ds2 = ds1[dict(time=0)]
    ds2["time"] = np.datetime64("2019-05-31T06", "ns")
    ds2 = ds2.expand_dims("time")

    ds3 = ds1[dict(time=0)]
    ds3["time"] = np.datetime64("2019-05-31T04", "ns")
    ds3 = ds3.expand_dims("time")

    ds = xr.concat([ds3, ds1[dict(time=[0])], ds2], "time")
    return MetDataset(ds)


def test_met_hashable(zero_like_da: xr.DataArray) -> None:
    mda = MetDataArray(zero_like_da)
    assert isinstance(mda.hash, str)


def test_met_values_property(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> None:
    """Test that met dataarray values are correctly lazy-loaded."""

    # this dataarray is already loaded
    mda = MetDataArray(zero_like_da)
    assert mda.in_memory
    assert mda.values is mda.data.values

    # this dataarray is not already loaded
    ds = xr.open_mfdataset(met_ecmwf_pl_path)
    mda = MetDataArray(ds["t"])
    assert not mda.in_memory
    assert mda.values is mda.data.values
    assert mda.in_memory


@pytest.mark.parametrize(
    "values, dimension",
    [
        ((-211, 53, 250, np.datetime64("2019-05-31T05:30")), 0),
        ((26, -112, 250, np.datetime64("2019-05-31T05:30")), 1),
        ((26, 53, 100, np.datetime64("2019-05-31T05:30")), 2),
        ((26, 53, 250, np.datetime64("2019-05-31T06:30")), 3),
    ],
)
def test_doubled_time_mds_interp_outside_grid(
    pl_path_to_doubled_time_mds: MetDataset,
    values: tuple[int, int, int, np.datetime64],
    dimension: int,
) -> None:
    mds = pl_path_to_doubled_time_mds
    variable = "q"
    mda = mds[variable]

    with pytest.raises(ValueError, match=f"out of bounds in dimension {dimension}"):
        mda.interpolate(*values, bounds_error=True)


# we get a scipy warning caused by division by zero
# this is an artifact of passing in waypoints agreeing with mda index
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_interpolate_wrap_antimeridian(sl_path_to_tripled_time_mds: MetDataset) -> None:
    ds = sl_path_to_tripled_time_mds
    mda = MetDataArray(ds["sp"].data, wrap_longitude=True)
    waypoints = pd.DataFrame(
        {
            "longitude": np.concatenate([np.linspace(-180, -160, 100), np.linspace(170, 180, 100)]),
            "latitude": 10,  # should be close to 15
            "level": -1,
            "time": pd.date_range("2019-05-31T04", "2019-05-31T05:59:59", 200),
        }
    )

    interpolated = mda.interpolate(
        waypoints["longitude"].values,
        waypoints["latitude"].values,
        waypoints["level"].values,
        waypoints["time"].values,
        method="nearest",
        bounds_error=True,
    )
    waypoints["interpolated"] = interpolated

    # mda['longitude'] = -160, -135, ..., 150, 175
    # waypoint longitudes < -172.5 are closest to mda longitude 175
    mda_value_175 = mda.data.sel(longitude=175, latitude=15).isel(time=0).values.item()
    mda_value_m160 = mda.data.sel(longitude=-160, latitude=15).isel(time=0).values.item()

    east_waypoints = waypoints["longitude"].between(-172.5, -160)
    mid_waypoints = waypoints["longitude"] < -172.5
    west_waypoints = waypoints["longitude"] > 170

    assert (waypoints.loc[east_waypoints, "interpolated"] == mda_value_m160).all()
    assert (waypoints.loc[mid_waypoints, "interpolated"] == mda_value_175).all()
    assert (waypoints.loc[west_waypoints, "interpolated"] == mda_value_175).all()


def test_interpolate_island_binary(island_binary: MetDataArray) -> None:
    # values in the island are 1
    t0, t1 = island_binary.data["time"].values[[0, -1]]
    waypoints = _generate_waypoints(-4, 4, -4, 4, 150, 350, t0, t1)
    interpolated = island_binary.interpolate(*waypoints)
    np.testing.assert_almost_equal(interpolated, np.ones((len(waypoints[0]),)))

    # and values outside are 0
    waypoints = _generate_waypoints(20, 30, 40, 50, 150, 350, t0, t1)
    interpolated = island_binary.interpolate(*waypoints)
    np.testing.assert_array_equal(interpolated, np.zeros((len(waypoints[0]),)))


def test_interpolate_antimeridian_binary(antimeridian_binary: MetDataArray) -> None:
    # values in the island are 1
    t0, t1 = antimeridian_binary.data["time"].values[[0, -1]]
    waypoints = _generate_waypoints(-180, -176, -4, 4, 150, 350, t0, t1)
    interpolated = antimeridian_binary.interpolate(*waypoints)
    np.testing.assert_almost_equal(interpolated, np.ones((len(waypoints[0]),)))

    # and values outside are 0
    waypoints = _generate_waypoints(-150, -140, -4, 4, 150, 350, t0, t1)
    interpolated = antimeridian_binary.interpolate(*waypoints)
    np.testing.assert_array_equal(interpolated, np.zeros((len(waypoints[0]),)))


def test_bad_interpolate(island_binary: MetDataArray) -> None:
    assert len(island_binary.interpolate(5, 5, 200, np.datetime64("2019-05-31T05:00:00"))) == 1

    with pytest.raises(ValueError):
        island_binary.interpolate(
            200, 5, 200, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
        )
    with pytest.raises(ValueError):
        island_binary.interpolate(
            5, 100, 200, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
        )
    with pytest.raises(ValueError):
        island_binary.interpolate(
            5, 5, 600, np.datetime64("2019-05-31T05:00:00"), bounds_error=True
        )
    with pytest.raises(ValueError):
        island_binary.interpolate(
            5, 5, 200, np.datetime64("2021-05-31T05:00:00"), bounds_error=True
        )


def test_binary(median_binary: MetDataArray) -> None:
    assert isinstance(median_binary, MetDataArray)
    assert isinstance(median_binary.proportion, float)
    assert 0.45 < median_binary.proportion < 0.55
    assert median_binary.binary


def test_edges_on_sparse_binary(sparse_binary: tuple[MetDataArray, dict]) -> None:
    mda, _ = sparse_binary
    edges = mda.find_edges()
    # edges is either all 0, or equals the original mda
    if edges.data.sum().item() > 0:
        assert (edges.data == mda.data).all().item()
    else:
        assert edges.data.sum().item() == 0


def test_edges_on_island_binary(island_binary: MetDataArray) -> None:
    edges = island_binary.find_edges()
    # 4 levels, 5 times, 4 sides to the island, 8 points to side
    assert edges.data.sum().item() == 4 * 5 * 4 * 8
    for i in range(-3, 4):
        for j in range(-3, 4):
            assert np.all(edges.data.sel(latitude=i, longitude=j) == 0)
    for i in range(-4, 5):
        for j in [-4, 4]:
            assert np.all(edges.data.sel(latitude=i, longitude=j) == 1)
            assert np.all(edges.data.sel(latitude=j, longitude=i) == 1)


def test_edges_on_antimeridian_binary(antimeridian_binary: MetDataArray) -> None:
    assert antimeridian_binary.shape == (361, 181, 4, 5)
    edges = antimeridian_binary.find_edges()
    assert edges.data.sum().item() == 4 * 5 * (4 * 9 + 6 * 2)  # additional boundary points added


def test_edges_on_median_binary(median_binary: MetDataArray) -> None:
    edges = median_binary.find_edges()

    # edges only exists on points with value 1
    assert ((median_binary.data - edges.data) >= 0).all().item()
    assert np.all(edges.data["longitude"] == median_binary.data["longitude"])  # consistent index
    assert np.all(edges.data["latitude"] == median_binary.data["latitude"])
    assert np.all(edges.data["level"] == median_binary.data["level"])
    assert np.all(edges.data["time"] == median_binary.data["time"])


def test_polygons_sparse_binary_specify_time_level(
    sparse_binary: tuple[MetDataArray, dict]
) -> None:
    mda, _ = sparse_binary
    # must specify level and time
    with pytest.raises(ValueError, match="coordinates is not 1"):
        mda.to_polygon_feature(iso_value=0.5)

    with pytest.raises(ValueError, match="coordinates is not 1"):
        mda.to_polygon_feature(level=mda.data["level"][0], iso_value=0.5)

    with pytest.raises(ValueError, match="coordinates is not 1"):
        mda.to_polygon_feature(time=mda.data["time"][0], iso_value=0.5)

    # if both are specified, we're okay
    geojson1 = mda.to_polygon_feature(
        level=mda.data["level"][0],
        time=mda.data["time"][0],
        iso_value=0.5,
    )

    # or if only a single level / time slice exists, we're okay
    # a single time coord
    sliced = MetDataArray(mda.data.isel(time=[0]))
    geojson2 = sliced.to_polygon_feature(level=mda.data["level"][0], iso_value=0.5)

    # a single level coord
    sliced = MetDataArray(mda.data.isel(level=[0]))
    geojson3 = sliced.to_polygon_feature(time=mda.data["time"][0], iso_value=0.5)

    # single time and level coords
    sliced = MetDataArray(mda.data.isel(time=[0], level=[0]))
    geojson4 = sliced.to_polygon_feature(iso_value=0.5)

    assert geojson1 == geojson2 == geojson3 == geojson4


@pytest.mark.parametrize("time", range(5))
def test_polygon_sparse_binary(
    sparse_binary: tuple[MetDataArray, dict], time: np.datetime64
) -> None:
    mda, nonzero_coord = sparse_binary
    mda = MetDataArray(mda.data.isel(time=[time]))
    level = nonzero_coord["level"]

    geojson = mda.to_polygon_feature(
        level=level,
        min_area=0.0,
        iso_value=0.5,
        epsilon=0.0,
        precision=6,
    )
    assert geojson["type"] == "Feature"
    assert geojson["geometry"]["type"] == "MultiPolygon"
    coords = geojson["geometry"]["coordinates"]
    assert isinstance(coords, list)

    # multipolygons are arrays of Polygons
    assert len(coords) == 1 and len(coords[0])
    coords = coords[0][0]

    # polygon closes in on itself
    assert coords[0] == coords[-1]

    # and has 4 distinct vertices
    assert len(coords) == 5
    coords = set(tuple(coord) for coord in coords)  # casting to set (an making each coord hashable)
    assert len(coords) == 4

    # and polygon successfully encloses nonzero point
    lon = nonzero_coord["lon"]
    lat = nonzero_coord["lat"]
    assert set(coord[0] for coord in coords) == {lon - 0.5, lon, lon + 0.5}
    assert set(coord[1] for coord in coords) == {lat - 0.5, lat, lat + 0.5}

    # polygon is JSON serializable
    assert json.dumps(geojson)


@pytest.fixture
def island_slice(island_binary: MetDataArray) -> MetDataArray:
    # every (level, time) slice of island_binary is identical -- just grab one
    return MetDataArray(island_binary.data.isel(level=[0], time=[0]))


def test_polygon_island_binary(island_slice: MetDataArray) -> None:
    geojson = island_slice.to_polygon_feature(iso_value=0.5, epsilon=0.01, precision=2)

    # contains a single polygon
    coords = geojson["geometry"]["coordinates"]
    assert isinstance(coords, list)
    assert len(coords) == 1 and len(coords[0]) == 1
    coords = coords[0][0]

    # polygon closes in on itself
    assert coords[0] == coords[-1]

    # polygon is square-shaped with the corners cut off
    # we have exactly two vertices at each island corner
    assert len(coords) == 9

    # all vertices have lon and lat value \pm 4 or \pm 4.5
    for coord in coords:
        for component in coord:
            assert component in [-4.5, -4, 4, 4.5]


@pytest.mark.parametrize("interiors", (True, False))
def test_nested_polygons(zero_like_da: xr.DataArray, interiors: bool) -> None:
    """This test shows that we *dont* currently support deeply nested polygons.

    To update once we support this feature.
    """
    da = zero_like_da.copy()

    island = (np.abs(da["latitude"]) < 50) & (np.abs(da["longitude"]) < 50)
    island2 = (np.abs(da["latitude"]) < 30) & (np.abs(da["longitude"]) < 30)
    island3 = (np.abs(da["latitude"]) < 20) & (np.abs(da["longitude"]) < 20)
    island4 = (np.abs(da["latitude"]) < 10) & (np.abs(da["longitude"]) < 10)
    island5 = (np.abs(da["latitude"]) < 5) & (np.abs(da["longitude"]) < 5)

    da = xr.where((island), 1, da)
    da = xr.where((island2), 0, da)
    da = xr.where((island3), 1, da)
    da = xr.where((island4), 0, da)
    da = xr.where((island5), 1, da)

    mda = MetDataArray(da.isel(level=[0], time=[0]))
    geojson = mda.to_polygon_feature(iso_value=0.5, interiors=interiors)

    # contains 1 Polygon with 1 nested polygon
    coords = geojson["geometry"]["coordinates"]
    assert isinstance(coords, list)

    if interiors:
        assert len(coords) == 3
    else:
        assert len(coords) == 1
        assert len(coords[0]) == 1


def test_polygons_with_holes():
    """Test ``to_polygon_feature`` method on handcrafted example.

    To get some intuition for this test, print or plot ``a`` in a notebook.
    """
    # Build up some data with interesting topology
    a = np.zeros((20, 10))

    # polygon 0: no holes
    a[1:11, 8] = 1

    # polygon 1: no holes
    a[2, 2] = 1
    a[2, 3] = 1
    a[3, 2] = 1

    # polygon 2: one hole
    a[8:12, 2:5] = 1
    a[9:11, 3] = 0

    # polygon 3: two holes
    a[13:17, 2:8] = 1
    a[14:16, 3:5] = 0
    a[15, 6] = 0
    a[13, 6:8] = 0

    # Pin the sum
    assert a.sum() == 40

    # Run through to_polygon_feature
    da = xr.DataArray(a, {"longitude": np.arange(20), "latitude": np.arange(10)})
    da = da.expand_dims(level=[-1], time=[0])
    mda = MetDataArray(da)
    polys = mda.to_polygon_feature(iso_value=0.5)["geometry"]["coordinates"]
    assert len(polys) == 4

    lens = [len(p) for p in polys]
    assert lens == [3, 2, 1, 1]

    # Pin some values from polygon 0
    outer, inner1, inner2 = polys[0]
    assert len(outer) == 12
    assert len(inner1) == 5  # octagon
    assert len(inner2) == 9  # diamond


@pytest.mark.parametrize("iso_value", [0, 0.001, 0.5, 0.9, 1])
def test_polygon_iso_value(island_slice: MetDataArray, iso_value: float) -> None:
    geojson = island_slice.to_polygon_feature(iso_value=iso_value, epsilon=0.1)
    coords = geojson["geometry"]["coordinates"]
    assert len(coords[0][0]) == 9


@pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
@pytest.mark.parametrize("return_type", ["geojson", "mesh"])
def test_polyhedra_save(median_binary: MetDataArray, return_type: str) -> None:
    assert len(median_binary.data["time"]) == 1
    path: str | pathlib.Path
    if return_type == "geojson":
        path = "test_mesh.json"
    else:
        path = "test_mesh.ply"
    median_binary.to_polyhedra(return_type=return_type, path=path)

    path = pathlib.Path(path)
    assert path.is_file()
    path.unlink()


@pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
def test_polyhedra_specify_time(sparse_binary: tuple[MetDataArray, dict]) -> None:
    mda, _ = sparse_binary
    with pytest.raises(ValueError, match="time input must be defined when the length"):
        mda.to_polyhedra()
    time = mda.data["time"][0]
    assert mda.to_polyhedra(time=time)


@pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
def test_polyhedra_mesh(sparse_binary: tuple[MetDataArray, dict]) -> None:
    """Test the `to_polyhedra` method with return_type = "mesh"."""
    mda, nonzero = sparse_binary
    mda = MetDataArray(mda.data.isel(time=[0]))

    mesh = mda.to_polyhedra(return_type="mesh")
    assert len(mesh.triangles) in [4, 8]
    if len(mesh.triangles) == 4:
        assert len(mesh.vertices) == 5
        assert not mesh.is_watertight()
    else:
        assert len(mesh.vertices) == 6
        assert mesh.is_watertight()

    lon, lat, _ = mesh.get_center()
    assert lon == nonzero["lon"]
    assert lat == nonzero["lat"]

    mesh2 = mda.to_polyhedra(return_type="mesh", closed=True)
    assert len(mesh2.triangles) == 8
    assert len(mesh2.vertices) == 6
    assert mesh2.is_watertight()


@pytest.mark.skipif(not OPEN3D_AVAILABLE, reason="Open3D not available")
def test_polyhedra_geojson(sparse_binary: tuple[MetDataArray, dict]) -> None:
    """Test the `to_polyhedra` method."""

    mda, _ = sparse_binary
    mda = MetDataArray(mda.data.isel(time=[0]))

    # TODO: this test was written where closed=False, should add a test for closed
    geojson = mda.to_polyhedra(closed=False)
    assert isinstance(geojson, dict)
    assert geojson["type"] == "Feature"
    assert geojson["geometry"]["type"] == "MultiPolygon"
    coords = geojson["geometry"]["coordinates"]
    assert len(coords) in [4, 8]

    # comprised of triangles
    assert all(len(face) == 3 for poly in coords for face in poly)

    # and each point has three components
    assert all(len(point) == 3 for poly in coords for face in poly for point in face)

    # and altitude values are reasonable
    mda_altitude = set(mda.data["altitude"].values.astype(int))
    geojson_altitude = set([int(point[2]) for poly in coords for face in poly for point in face])
    assert geojson_altitude.issubset(mda_altitude)


def test_save_load(met_ecmwf_pl_path: str, met_era5_fake: MetDataset) -> None:
    """Test the `save` and `load` methods."""
    _cache = DiskCacheStore(cache_dir=f"{DISK_CACHE_DIR}/test", allow_clear=True)
    _cache.clear()
    mds = MetDataset(xr.open_dataset(met_ecmwf_pl_path), cachestore=_cache)

    # add new field
    mds.data["new"] = xr.ones_like(mds.data["t"])

    # save to cache
    assert not _cache.exists(f"{mds.hash}-0.nc")
    mds.save()
    assert _cache.exists(f"{mds.hash}-0.nc")

    mds2 = MetDataset.load(mds.hash, cachestore=_cache)
    assert "new" in mds2 and np.all(mds2.data["new"].values == 1)

    # met data array version
    mda = MetDataArray(mds.data["t"], cachestore=_cache)

    # save to cache
    assert not _cache.exists(f"{mda.hash}-0.nc")
    mda.save()
    assert _cache.exists(f"{mda.hash}-0.nc")

    mda2 = MetDataArray.load(mda.hash, cachestore=_cache)
    assert mda2.name == mda.name
    xr.testing.assert_equal(mda.data, mda2.data)

    # save multiple time slices separately - dataset
    mds = met_era5_fake.copy()
    mds.cachestore = _cache
    assert not _cache.exists(f"{mds.hash}-0.nc") and not _cache.exists(f"{mds.hash}-1.nc")
    mds.save()

    assert _cache.exists(f"{mds.hash}-0.nc")
    assert _cache.exists(f"{mds.hash}-1.nc")
    assert _cache.exists(f"{mds.hash}-2.nc")

    mds2 = MetDataset.load(mds.hash, cachestore=_cache)
    assert np.all(mds2.data["air_temperature"].values == 230)
    assert len(mds2.data["time"]) == 4

    # save multiple time slices separately - dataarray
    mda = MetDataArray(met_era5_fake.data["air_temperature"])
    mda.cachestore = _cache
    assert mda.hash != mds.hash
    assert not _cache.exists(f"{mda.hash}-0.nc") and not _cache.exists(f"{mda.hash}-1.nc")
    mda.save()

    assert _cache.exists(f"{mda.hash}-0.nc")
    assert _cache.exists(f"{mda.hash}-1.nc")
    assert _cache.exists(f"{mda.hash}-2.nc")

    mda2 = MetDataArray.load(mda.hash, cachestore=_cache)
    assert np.all(mda2.data.values == 230)
    assert len(mda2.data["time"]) == 4

    # cleanup
    _cache.clear()


def test_met_size(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> None:
    """Test `MetBase.size` property."""

    mda = MetDataArray(zero_like_da)
    assert mda.size == 360 * 181 * 4 * 5

    mds = MetDataset(xr.open_dataset(met_ecmwf_pl_path))
    assert mds.size == 8 * 15 * 3 * 2


def test_met_copy(met_ecmwf_pl_path: str) -> None:
    """Check that the method `copy` works as expected with a silly numerical experiment.."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds)
    mds2 = mds.copy()

    assert mds.data is not mds2.data

    assert round(mds["t"].data[5][5][2][0].item()) == 235
    mds["t"].data *= 2
    assert round(mds["t"].data[5][5][2][0].item()) == 470
    assert round(mds2["t"].data[5][5][2][0].item()) == 235


def test_met_wrap_longitude_chunks(met_ecmwf_pl_path: str, override_cache: DiskCacheStore) -> None:
    """Check that the wrap_longitude method increments longitudinal chunks."""
    # Using the xr.open_dataset method
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds)

    # NOTE: mds doesn't have chunks
    assert mds.data.chunks == {}
    wrapped = mds.wrap_longitude()
    assert wrapped.data.chunks == {}
    assert len(wrapped.data["longitude"]) == len(mds.data["longitude"]) + 2

    # Using our open_metdataset method
    era5 = ERA5(
        time=ds["time"].values,
        variables=list(ds.data_vars),
        pressure_levels=ds["level"].values,
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    mds2 = era5.open_metdataset()

    # This one's got chunks
    chunks = mds2.data.chunks
    assert "longitude" in chunks
    assert chunks["longitude"] == (15,)

    wrapped2 = mds2.wrap_longitude()
    assert len(wrapped.data["longitude"]) == len(mds.data["longitude"]) + 2

    # successfully incremented chunks
    assert wrapped2.data.chunks["longitude"] == (17,)


def test_broadcast_coords(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> None:
    """Test the `broadcast_coords` method."""

    mda = MetDataArray(zero_like_da)
    da1 = mda.broadcast_coords("level")
    assert da1.name == "level"
    assert da1.shape == (360, 181, 4, 5)
    assert da1[0, 1, 1, 1] == 200

    # non-dimension coords
    da2 = mda.broadcast_coords("air_pressure")
    assert da2.name == "air_pressure"
    assert da2.shape == (360, 181, 4, 5)
    assert da2[0, 1, 1, 1] == 20000

    # met dataset
    mds = MetDataset(xr.open_dataset(met_ecmwf_pl_path))
    da3 = mds.broadcast_coords("longitude")
    assert da3.name == "longitude"
    assert da3.shape == (15, 8, 3, 2)
    assert da3[0, 1, 1, 0] == -160


def test_downselect(zero_like_da: xr.DataArray, met_ecmwf_pl_path: str) -> None:
    """Test the `downselect` method."""

    mda = MetDataArray(zero_like_da)
    mda2 = mda.downselect([0, 0, 10, 20])
    assert mda is not mda2  # returns a copy
    assert mda2.name == mda.name
    assert (mda2.data["longitude"].min(), mda2.data["longitude"].max()) == (0, 10)
    assert (mda2.data["latitude"].min(), mda2.data["latitude"].max()) == (0, 20)
    assert (mda2.data["level"].min(), mda2.data["level"].max()) == (
        mda.data["level"].min(),
        mda.data["level"].max(),
    )

    # level selection
    mda2 = mda.downselect([0, 0, 200, 10, 20, 300])
    assert (mda2.data["longitude"].min(), mda2.data["longitude"].max()) == (0, 10)
    assert (mda2.data["latitude"].min(), mda2.data["latitude"].max()) == (0, 20)
    assert (mda2.data["level"].min(), mda2.data["level"].max()) == (200, 300)

    with pytest.raises(ValueError, match="is not length 4"):
        mda.downselect([0, 0, 200, 10, 20, 300, 5000])

    # wrap data
    mda2 = mda.downselect([179, 0, -179, 20])
    assert (mda2.data["longitude"].min(), mda2.data["longitude"].max()) == (-180, 179)
    assert np.all(mda2.data["longitude"] == np.array([-180, -179, 179]))
    assert (mda2.data["latitude"].min(), mda2.data["latitude"].max()) == (0, 20)

    # met dataset
    mds = MetDataset(xr.open_dataset(met_ecmwf_pl_path))
    mds2 = mds.downselect([0, 0, 125, 40])
    assert mds is not mds2  # does not copy
    assert (mds2.data["longitude"].min(), mds2.data["longitude"].max()) == (0, 125)
    assert (mds2.data["latitude"].min(), mds2.data["latitude"].max()) == (15, 40)
    assert (mds2.data["level"].min(), mds2.data["level"].max()) == (
        mds.data["level"].min(),
        mds.data["level"].max(),
    )

    mds2 = mds.downselect([0, 0, 200, 125, 40, 300])
    assert (mds2.data["longitude"].min(), mds2.data["longitude"].max()) == (0, 125)
    assert (mds2.data["latitude"].min(), mds2.data["latitude"].max()) == (15, 40)
    assert (mds2.data["level"].min(), mds2.data["level"].max()) == (225, 300)


def test_met_dataset_iter(met_ecmwf_pl_path: str) -> None:
    """Test implementation of MetDataset.__iter__."""
    met = MetDataset(xr.open_dataset(met_ecmwf_pl_path))
    assert list(met) == ["t", "r", "q", "ciwc"]


def test_met_dataset_setitem(met_ecmwf_pl_path: str) -> None:
    """Test implementation of MetDataset.__setitem__."""
    met = MetDataset(xr.open_dataset(met_ecmwf_pl_path))

    # both of the following should work
    met.data["r2"] = met.data["t"].copy()
    met["r3"] = met["t"].copy()
    np.testing.assert_array_equal(met["r2"].values, met["t"].values)
    np.testing.assert_array_equal(met["r3"].values, met["t"].values)

    met.data["r2"] = met["t"].copy().data
    np.testing.assert_array_equal(met["r2"].values, met["t"].values)

    with pytest.warns(UserWarning, match=r"Overwriting data in keys `\['r'\]`"):
        met["r"] = met["t"].copy()

    # update should work without warnings
    met.data.update({"r2": met.data["t"].copy()})
    np.testing.assert_array_equal(met["r2"].values, met["t"].values)

    met.update({"r3": met["t"].copy()})
    np.testing.assert_array_equal(met["r3"].values, met["t"].values)
    met.update({"r4": met.data["t"].copy()})
    np.testing.assert_array_equal(met["r4"].values, met["t"].values)

    met.data.update({"r5": met["t"].copy().data})
    np.testing.assert_array_equal(met["r5"].values, met["t"].values)


def test_met_variable() -> None:
    """Test MetVariable dataclass."""

    # new parameters
    NewParam = MetVariable(ecmwf_id=829, standard_name="new-param", short_name="wt", units="n")
    assert isinstance(NewParam, MetVariable)
    assert NewParam.short_name == "wt"
    assert NewParam.ecmwf_link is not None and "grib/param-db?id=829" in NewParam.ecmwf_link

    NewParam2 = MetVariable(standard_name="new-param2", short_name="wt", units="n")
    assert NewParam2.ecmwf_link is None

    # standard name and short name are required
    with pytest.raises(TypeError, match="short_name"):
        MetVariable(standard_name="new-param")

    with pytest.raises(TypeError, match="standard_name"):
        MetVariable(short_name="new-param")

    # should have attrs property
    NewParam = MetVariable(ecmwf_id=829, standard_name="new-param", short_name="wt", units="n")
    attrs = NewParam.attrs
    assert "standard_name" in attrs and attrs["standard_name"] == "new-param"
    assert "short_name" in attrs and attrs["short_name"] == "wt"
    assert "long_name" not in attrs
    assert "ecmwf_id" not in attrs


def test_met_ensure_vars(met_ecmwf_pl_path: str) -> None:
    """Test `met.ensure_vars()."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    met = MetDataset(ds)

    # should work with strings and return string names
    variables = met.ensure_vars(["t", "r", "q"])
    assert variables == ["t", "r", "q"]

    with pytest.raises(KeyError, match="Dataset does not contain variable `w`"):
        met.ensure_vars(["t", "r", "w"])

    # should work with MetVariables and return standard_name
    t = MetVariable(short_name="t", standard_name="t")
    r = MetVariable(short_name="r", standard_name="r")
    w = MetVariable(short_name="w", standard_name="w")
    variables = met.ensure_vars([t, r])
    assert variables == ["t", "r"]

    with pytest.raises(KeyError, match="Dataset does not contain variable `w`"):
        met.ensure_vars([t, r, w])

    # should work with a list of MetVariables and return standard_name
    variables = met.ensure_vars([[t, w], r])
    assert variables == ["t", "r"]

    variables = met.ensure_vars(["r", [w, t]])
    assert variables == ["r", "t"]


@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
def test_include_altitude(met_ecmwf_pl_path: str):
    """Check `to_polygon_feature` with `include_altitude=True`."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds[dict(time=[0])])
    mda = mds["t"]

    geojson = mda.to_polygon_feature(level=225, include_altitude=True, iso_value=100)
    for poly in geojson["geometry"]["coordinates"]:
        for ring in poly:
            for coord in ring:
                assert len(coord) == 3
                assert coord[2] == 11037.0


@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
def test_include_altitude_raises(met_ecmwf_sl_path: str):
    """Check `to_polygon_feature` raises a ValueError when `include_altitude=True` and SL data."""
    ds = xr.open_dataset(met_ecmwf_sl_path)
    mds = MetDataset(ds[dict(time=[0])])
    mda = mds["sp"]
    with pytest.raises(ValueError, match=" but altitude is not found"):
        mda.to_polygon_feature(include_altitude=True, iso_value=20000)

    geojson = mda.to_polygon_feature(include_altitude=False, iso_value=20000)
    for poly in geojson["geometry"]["coordinates"]:
        for ring in poly:
            for coord in ring:
                assert len(coord) == 2


@pytest.mark.filterwarnings("ignore:Longitude is not evenly spaced")
def test_to_polygon_feature_collection(met_ecmwf_pl_path: str):
    """Confirm that `to_polygon_feature_collection` works."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds[dict(time=[0])])
    mda = mds["q"]

    fc = mda.to_polygon_feature_collection(iso_value=0.0001)
    assert fc.keys() == {"type", "features"}
    assert fc["type"] == "FeatureCollection"
    features = fc["features"]
    assert len(features) == len(ds.level)
    for feature in features:
        assert feature["properties"]


def test_wrap_longitude_new_patterns():
    """Test new wrap_longitude patterns (pycontails 0.26.0)."""
    da1 = xr.DataArray(
        data=np.random.random((360, 5, 5, 1)),
        dims=["longitude", "latitude", "level", "time"],
        coords={
            "longitude": np.arange(-180, 180),
            "latitude": np.arange(0, 50, 10),
            "level": np.arange(50, 500, 100),
            "time": [datetime(2000, 1, 1, 0)],
        },
    )
    # Includes new longitude at 180
    mda1 = MetDataArray(da1, wrap_longitude=True)
    assert mda1.shape == (361, 5, 5, 1)
    assert mda1.is_wrapped
    np.testing.assert_array_equal(mda1.data.longitude, np.arange(-180, 181))
    np.testing.assert_array_equal(mda1.data.sel(longitude=180), mda1.data.sel(longitude=-180))

    # Drop longitude at -180
    da2 = da1.isel(longitude=slice(1, None))
    assert da2.shape == (359, 5, 5, 1)
    np.testing.assert_array_equal(da2.longitude, np.arange(-179, 180, 1))

    # New mda includes value at -181 and value at 181
    mda2 = MetDataArray(da2, wrap_longitude=True)
    assert mda2.shape == (361, 5, 5, 1)
    assert mda2.is_wrapped
    np.testing.assert_array_equal(mda2.data.longitude, np.r_[-181, np.arange(-179, 180), 181])

    # Drop longitude at 179
    da3 = da1.isel(longitude=slice(None, -1))
    assert da3.shape == (359, 5, 5, 1)
    np.testing.assert_array_equal(da3.longitude, np.arange(-180, 179, 1))
    mda3 = MetDataArray(da3, wrap_longitude=True)
    assert mda3.shape == (360, 5, 5, 1)
    assert mda3.is_wrapped
    np.testing.assert_array_equal(mda3.data.longitude, np.r_[np.arange(-180, 179), 180])


def test_originates_from_ecmwf_met_ecmwf_pl_path(met_ecmwf_pl_path: str):
    """Test the `originates_from_ecmwf` property."""
    ds = xr.open_dataset(met_ecmwf_pl_path)
    mds = MetDataset(ds)
    assert originates_from_ecmwf(mds)


def test_originates_from_ecmwf_met_ecmwf_sl_path(met_ecmwf_sl_path: str):
    """Test the `originates_from_ecmwf` property."""
    ds = xr.open_dataset(met_ecmwf_sl_path)
    mds = MetDataset(ds)
    assert originates_from_ecmwf(mds)


def test_originates_from_ecmwf_met_pcc_pl(met_pcc_pl: MetDataset):
    """Test the `originates_from_ecmwf` property."""
    assert originates_from_ecmwf(met_pcc_pl)


def test_originates_from_ecmwf_sparse_binary(sparse_binary: tuple[MetDataArray, Any]):
    """Test the `originates_from_ecmwf` property."""
    da = sparse_binary[0].data
    ds = da.to_dataset()
    mds = MetDataset(ds)
    assert not originates_from_ecmwf(mds)



"""
Module for calculating radiative forcing of contrail cirrus.

References
----------
- :cite:`schumannEffectiveRadiusIce2011`
- :cite:`schumannParametricRadiativeForcing2012`
"""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt


@dataclass
class RFConstants:
    """
    Constants that are used to calculate the local contrail radiative forcing.

    See Table 1 of :cite:`schumannParametricRadiativeForcing2012`.

    Each coefficient has 8 elements, one corresponding to each contrail ice particle habit (shape)::

        [
            Sphere,
            Solid column,
            Hollow column,
            Rough aggregate,
            Rosette-6,
            Plate,
            Droxtal,
            Myhre,
        ]

    For each waypoint, the distinct mix of ice particle habits are approximated using the mean
    contrail ice particle radius (``r_vol_um``) relative to ``radius_threshold_um``.

    For example:

    - if ``r_vol_um`` for a waypoint < 5 um, the mix of ice particle habits will be 100% droxtals.
    - if ``r_vol_um`` for a waypoint between 5 and 9.5 um, the mix of ice particle habits will
      be 30% solid columns, 70% droxtals.

    See Table 2 from :cite:`schumannEffectiveRadiusIce2011`.


    References
    ----------
    - :cite:`schumannEffectiveRadiusIce2011`
    - :cite:`schumannParametricRadiativeForcing2012`
    """

    # -----
    # Variables/coefficients used to calculate the local contrail longwave radiative forcing.
    # -----

    #: Linear approximation of Stefan-Boltzmann Law
    #: :math:`k_t` in Eq. (2) in :cite:`schumannParametricRadiativeForcing2012`
    k_t = np.array([1.93466, 1.95456, 1.95994, 1.95906, 1.94397, 1.95123, 2.30363, 1.94611])

    #: Approximates the temperature of the atmosphere without contrails
    #: :math:`T_{0}` in Eq. (2) in :cite:`schumannParametricRadiativeForcing2012`
    T_0 = np.array([152.237, 152.724, 152.923, 152.360, 151.879, 152.318, 165.692, 153.073])

    #: Approximate the effective emmissivity factor
    #: :math:`\delta_{\tau} in :cite:`schumannParametricRadiativeForcing2012`
    delta_t = np.array(
        [0.940846, 0.808397, 0.736222, 0.675591, 0.748757, 0.708515, 0.927592, 0.795527]
    )

    #: Effective radius scaling factor for optical properties (extinction relative to scattering)
    #: :math:`\delta_{lr} in Eq. (3) in :cite:`schumannParametricRadiativeForcing2012`
    delta_lr = np.array([0.211276, 0.341194, 0.325496, 0.255921, 0.170265, 1.65441, 0.201949, 0])

    #: Optical depth scaling factor for reduction of the OLR at the contrail level due
    #: to existing cirrus above the contrail
    #: :math:`\delta_{lc} in Eq. (4) in :cite:`schumannParametricRadiativeForcing2012`
    delta_lc = np.array(
        [0.159942, 0.0958129, 0.0924850, 0.0462023, 0.132925, 0.0870067, 0.0626339, 0.0665289]
    )

    # -----
    # Variables/coefficients used to calculate the local contrail shortwave radiative forcing.
    # -----

    #: Approximates the dependence on the effective albedo
    #: :math:`t_a`: Eq. (5) in :cite:`schumannParametricRadiativeForcing2012`
    t_a = np.array([0.879119, 0.901701, 0.881812, 0.899144, 0.879896, 0.883212, 0.899096, 1.00744])

    # Approximates the albedo of the contrail
    #: :math:`A_{\mu}` in Eq. (6) in :cite:`schumannParametricRadiativeForcing2012`
    A_mu = np.array(
        [0.361226, 0.294072, 0.343894, 0.317866, 0.337227, 0.310978, 0.342593, 0.269179]
    )

    # Approximates the albedo of the contrail
    #: :math:`C_{\mu}` in Eq. (6) in :cite:`schumannParametricRadiativeForcing2012`
    C_mu = np.array(
        [0.709300, 0.678016, 0.687546, 0.675315, 0.712041, 0.713317, 0.660267, 0.545716]
    )

    #: Approximates the effective contrail optical depth
    #: :math:`delta_sr` in Eq. (7) and (8) in :cite:`schumannParametricRadiativeForcing2012`
    delta_sr = np.array(
        [0.149851, 0.0254270, 0.0238836, 0.0463724, 0.0478892, 0.0700234, 0.0517942, 0]
    )

    #: Approximates the effective contrail optical depth
    #: :math:`F_r` in Eq. (7) and (8) in :cite:`schumannParametricRadiativeForcing2012`
    F_r = np.array([0.511852, 0.576911, 0.597351, 0.225750, 0.550734, 0.817858, 0.249004, 0])

    #: Approximates the contrail reflectances
    #: :math:`\gamma` in Eq. (9) in :cite:`schumannParametricRadiativeForcing2012`
    gamma_lower = np.array(
        [0.323166, 0.392598, 0.356189, 0.345040, 0.407515, 0.523604, 0.310853, 0.274741]
    )

    #: Approximates the contrail reflectances
    #: :math:`\Gamma` in Eq. (9) in :cite:`schumannParametricRadiativeForcing2012`
    gamma_upper = np.array(
        [0.241507, 0.347023, 0.288452, 0.296813, 0.327857, 0.437560, 0.274710, 0.208154]
    )

    #: Approximate the SZA-dependent contrail sideward scattering
    #: :math:`B_{\mu}` in Eq. (10) in :cite:`schumannParametricRadiativeForcing2012`
    B_mu = np.array([1.67592, 1.55687, 1.71065, 1.55843, 1.70782, 1.71789, 1.56399, 1.59015])

    #: Account for the optical depth of natural cirrus above the contrail
    #: :math:`\delta_{sc}` in Eq. (11) in :cite:`schumannParametricRadiativeForcing2012`
    delta_sc = np.array(
        [0.157017, 0.143274, 0.167995, 0.148547, 0.173036, 0.162442, 0.171855, 0.213488]
    )

    #: Account for the optical depth of natural cirrus above the contrail
    # :math:`\delta'_{sc}` in Eq. (11) in :cite:`schumannParametricRadiativeForcing2012`
    delta_sc_aps = np.array(
        [0.229574, 0.197611, 0.245036, 0.204875, 0.248328, 0.254029, 0.244051, 0.302246]
    )


# create a new constants class to use within module
rf_const = RFConstants()


# ----------
# Ice Habits
# ----------


def habit_weights(
    r_vol_um: npt.NDArray[np.float_],
    habit_distributions: npt.NDArray[np.float_],
    radius_threshold_um: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""Assign weights to different ice particle habits for each waypoint.

    For each waypoint, the distinct mix of ice particle habits are approximated
    using the mean contrail ice particle radius (``r_vol_um``) binned by ``radius_threshold_um``.

    For example:

    - For waypoints with r_vol_um < 5 um, the mix of ice particle habits will
      be from Group 1 (100% Droxtals,
        refer to :attr:`CocipParams().habit_distributions`).
    - For waypoints with 5 um <= ``r_vol_um`` < 9.5 um, the mix of ice particle
      habits will be from Group 2 (30% solid columns, 70% droxtals)

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]
    habit_distributions : npt.NDArray[np.float_]
        Habit weight distributions.
        See :attr:`CocipParams().habit_distributions`
    radius_threshold_um : npt.NDArray[np.float_]
        Radius thresholds for habit distributions.
        See :attr:`CocipParams.radius_threshold_um`

    Returns
    -------
    npt.NDArray[np.float_]
        Array with shape ``n_waypoints x 8 columns``, where each column is the weights to the ice
        particle habits, [:math:`[0 - 1]`], and the sum of each column should be equal to 1.

    Raises
    ------
    ValueError
        Raises when ``habit_distributions`` do not sum to 1 across columns or
        if there is a size mistmatch with ``radius_threshold_um``.
    """
    # all rows of the habit weights should sum to 1
    if not np.all(np.round(np.sum(habit_distributions, axis=1), 3) == 1):
        raise ValueError("Habit weight distributions must sum to 1 across columns")

    if habit_distributions.shape[0] != (radius_threshold_um.size + 1):
        raise ValueError(
            "The number of rows in `habit_distributions` must equal 1 + the "
            "size of `radius_threshold_um`"
        )

    # assign ice particle habits for each waypoint
    idx = habit_weight_regime_idx(r_vol_um, radius_threshold_um)
    return habit_distributions[idx]


def habit_weight_regime_idx(
    r_vol_um: npt.NDArray[np.float_], radius_threshold_um: npt.NDArray[np.float_]
) -> npt.NDArray[np.intp]:
    r"""
    Determine regime of ice particle habits based on contrail ice particle volume mean radius.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]
    radius_threshold_um : npt.NDArray[np.float_]
        Radius thresholds for habit distributions.
        See :attr:`CocipParams.radius_threshold_um`

    Returns
    -------
    npt.NDArray[np.intp]
        Row index of the habit distribution in array :attr:`CocipParams().habit_distributions`
    """
    # find the regime for each waypoint using thresholds
    idx = np.digitize(r_vol_um, radius_threshold_um)

    # set any nan values to the "0" type
    idx[np.isnan(r_vol_um)] = 0

    return idx


def effective_radius_by_habit(
    r_vol_um: npt.NDArray[np.float_], habit_idx: npt.NDArray[np.intp]
) -> np.ndarray:
    r"""Calculate the effective radius ``r_eff_um`` via the mean ice particle radius and habit type.

    The ``habit_idx`` corresponds to the habit types in ``rf_const.habits``.
    Each habit type has a specific parameterization to calculate ``r_eff_um`` based on ``r_vol_um``.
    derived from :cite:`schumannEffectiveRadiusIce2011`.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]
    habit_idx : npt.NDArray[np.intp]
        Habit type index for the contrail ice particle, corresponding to the
        habits in ``rf_const.habits``.

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius of ice particles for each combination of ``r_vol_um``
        and ``habit_idx``, [:math:`\mu m`]

    References
    ----------
    - :cite:`schumannEffectiveRadiusIce2011`
    """
    cond_list = [
        habit_idx == 0,
        habit_idx == 1,
        habit_idx == 2,
        habit_idx == 3,
        habit_idx == 4,
        habit_idx == 5,
        habit_idx == 6,
        habit_idx == 7,
    ]
    func_list = [
        effective_radius_sphere,
        effective_radius_solid_column,
        effective_radius_hollow_column,
        effective_radius_rough_aggregate,
        effective_radius_rosette,
        effective_radius_plate,
        effective_radius_droxtal,
        effective_radius_myhre,
        0.0,
    ]
    return np.piecewise(r_vol_um, cond_list, func_list)


def effective_radius_sphere(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a sphere particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    return np.minimum(r_vol_um, 25.0)


def effective_radius_solid_column(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a solid column particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = (
        0.2588 * np.exp(-(6.912e-3 * r_vol_um)) + 0.6372 * np.exp(-(3.142e-4 * r_vol_um))
    ) * r_vol_um
    is_small = r_vol_um <= 42.2
    r_eff_um[is_small] = 0.824 * r_vol_um[is_small]
    return np.minimum(r_eff_um, 45.0)


def effective_radius_hollow_column(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""Calculate the effective radius of ice particles assuming a hollow column particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = (
        0.2281 * np.exp(-(7.359e-3 * r_vol_um)) + 0.5651 * np.exp(-(3.350e-4 * r_vol_um))
    ) * r_vol_um
    is_small = r_vol_um <= 39.7
    r_eff_um[is_small] = 0.729 * r_vol_um[is_small]
    return np.minimum(r_eff_um, 45.0)


def effective_radius_rough_aggregate(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""Calculate the effective radius of ice particles assuming a rough aggregate particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = 0.574 * r_vol_um
    return np.minimum(r_eff_um, 45.0)


def effective_radius_rosette(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a rosette particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = r_vol_um * (
        0.1770 * np.exp(-(2.144e-2 * r_vol_um)) + 0.4267 * np.exp(-(3.562e-4 * r_vol_um))
    )
    return np.minimum(r_eff_um, 45.0)


def effective_radius_plate(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a plate particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = r_vol_um * (
        0.1663 + 0.3713 * np.exp(-(0.0336 * r_vol_um)) + 0.3309 * np.exp(-(0.0035 * r_vol_um))
    )
    return np.minimum(r_eff_um, 45.0)


def effective_radius_droxtal(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a droxtal particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    r_eff_um = 0.94 * r_vol_um
    return np.minimum(r_eff_um, 45.0)


def effective_radius_myhre(r_vol_um: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective radius of contrail ice particles assuming a sphere particle habit.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective radius, [:math:`\mu m`]
    """
    return np.minimum(r_vol_um, 45.0)


# -----------------
# Radiative Forcing
# -----------------


def longwave_radiative_forcing(
    r_vol_um: npt.NDArray[np.float_],
    olr: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    habit_weights_: npt.NDArray[np.float_],
    r_eff_um: npt.NDArray[np.float_] | None = None,
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the local contrail longwave radiative forcing (:math:`RF_{LW}`).

    All returned values are positive.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]
    olr : npt.NDArray[np.float_]
        Outgoing longwave radiation at each waypoint, [:math:`W m^{-2}`]
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature at each waypoint, [:math:`K`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth at each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the
        contrail at each waypoint
    habit_weights_ : npt.NDArray[np.float_]
        Weights to different ice particle habits for each waypoint,
        ``n_waypoints x 8`` (habit) columns, [:math:`[0 - 1]`]
    r_eff_um : npt.NDArray[np.float_], optional
        Provide effective radius corresponding to elements in ``r_vol_um``, [:math:`\mu m`].
        Defaults to None, which means the effective radius will be calculated using ``r_vol_um``
        and habit types in :func:`effective_radius_by_habit`.

    Returns
    -------
    npt.NDArray[np.float_]
        Local contrail longwave radiative forcing (positive), [:math:`W m^{-2}`]

    Raises
    ------
    ValueError
        If `r_eff_um` and `olr` have different shapes.

    References
    ----------
    - :cite:`schumannParametricRadiativeForcing2012`
    """
    # get list of habit weight indexs where the weights > 0
    # this is a tuple of (np.array[waypoint index], np.array[habit type index])
    habit_weight_mask = habit_weights_ > 0.0
    idx0, idx1 = np.nonzero(habit_weight_mask)

    # Convert parametric coefficients for vectorized operations
    delta_t = rf_const.delta_t[idx1]
    delta_lc = rf_const.delta_lc[idx1]
    delta_lr = rf_const.delta_lr[idx1]
    k_t = rf_const.k_t[idx1]
    T_0 = rf_const.T_0[idx1]

    olr_h = olr[idx0]
    tau_cirrus_h = tau_cirrus[idx0]
    tau_contrail_h = tau_contrail[idx0]
    air_temperature_h = air_temperature[idx0]

    # effective radius
    if r_eff_um is None:
        r_vol_um_h = r_vol_um[idx0]
        r_eff_um_h = effective_radius_by_habit(r_vol_um_h, idx1)
    else:
        if not isinstance(r_eff_um, np.ndarray) or r_eff_um.shape != olr.shape:
            raise ValueError(
                "User provided effective radius (`r_eff_um`) must have the same shape as `olr`"
                f" {olr.shape}"
            )

        r_eff_um_h = r_eff_um[idx0]

    # Longwave radiation calculations
    e_lw = olr_reduction_natural_cirrus(tau_cirrus_h, delta_lc)
    f_lw = contrail_effective_emissivity(r_eff_um_h, delta_lr)

    # calculate the RF LW per habit type
    # see eqn (2) in :cite:`schumannParametricRadiativeForcing2012`
    rf_lw_per_habit = (
        (olr_h - k_t * (air_temperature_h - T_0))
        * e_lw
        * (1.0 - np.exp(-delta_t * f_lw * tau_contrail_h))
    )
    rf_lw_per_habit = np.maximum(rf_lw_per_habit, 0.0)

    # Weight and sum the RF contributions of each habit type according the habit weight
    # regime at the waypoint
    # see eqn (12) in :cite:`schumannParametricRadiativeForcing2012`
    # use fancy indexing to re-assign values to 2d array of waypoint x habit type
    rf_lw_weighted = np.zeros_like(habit_weights_)
    rf_lw_weighted[idx0, idx1] = rf_lw_per_habit * habit_weights_[habit_weight_mask]
    return np.sum(rf_lw_weighted, axis=1)


def shortwave_radiative_forcing(
    r_vol_um: npt.NDArray[np.float_],
    sdr: npt.NDArray[np.float_],
    rsr: npt.NDArray[np.float_],
    sd0: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    habit_weights_: npt.NDArray[np.float_],
    r_eff_um: npt.NDArray[np.float_] | None = None,
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the local contrail shortwave radiative forcing (:math:`RF_{SW}`).

    All returned values are negative.

    Parameters
    ----------
    r_vol_um : npt.NDArray[np.float_]
        Contrail ice particle volume mean radius, [:math:`\mu m`]
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]
    sd0 : npt.NDArray[np.float_]
        Solar constant, [:math:`W m^{-2}`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the
        contrail for each waypoint.
    habit_weights_ : npt.NDArray[np.float_]
        Weights to different ice particle habits for each waypoint,
        ``n_waypoints x 8`` (habit) columns, [:math:`[0 - 1]`]
    r_eff_um : npt.NDArray[np.float_], optional
        Provide effective radius corresponding to elements in ``r_vol_um``, [:math:`\mu m`].
        Defaults to None, which means the effective radius will be calculated using ``r_vol_um``
        and habit types in :func:`effective_radius_by_habit`.

    Returns
    -------
    npt.NDArray[np.float_]
        Local contrail shortwave radiative forcing (negative), [:math:`W m^{-2}`]

    Raises
    ------
    ValueError
        If `r_eff_um` and `sdr` have different shapes.

    References
    ----------
    - :cite:`schumannParametricRadiativeForcing2012`
    """
    # create mask for daytime (sdr > 0)
    day = sdr > 0.0

    # short circuit if no waypoints occur during the day
    if not day.any():
        return np.zeros_like(sdr)

    # get list of habit weight indexs where the weights > 0
    # this is a tuple of (np.array[waypoint index], np.array[habit type index])
    habit_weight_mask = day.reshape(day.size, 1) & (habit_weights_ > 0.0)
    idx0, idx1 = np.nonzero(habit_weight_mask)

    # Convert parametric coefficients for vectorized operations
    t_a = rf_const.t_a[idx1]
    A_mu = rf_const.A_mu[idx1]
    B_mu = rf_const.B_mu[idx1]
    C_mu = rf_const.C_mu[idx1]
    delta_sr = rf_const.delta_sr[idx1]
    F_r = rf_const.F_r[idx1]
    gamma_lower = rf_const.gamma_lower[idx1]
    gamma_upper = rf_const.gamma_upper[idx1]
    delta_sc = rf_const.delta_sc[idx1]
    delta_sc_aps = rf_const.delta_sc_aps[idx1]

    sdr_h = sdr[idx0]
    rsr_h = rsr[idx0]
    sd0_h = sd0[idx0]
    tau_contrail_h = tau_contrail[idx0]
    tau_cirrus_h = tau_cirrus[idx0]

    albedo_ = albedo(sdr_h, rsr_h)
    mue = np.minimum(sdr_h / sd0_h, 1.0)

    # effective radius
    if r_eff_um is None:
        r_vol_um_h = r_vol_um[idx0]
        r_eff_um_h = effective_radius_by_habit(r_vol_um_h, idx1)
    else:
        if not isinstance(r_eff_um, np.ndarray) or r_eff_um.shape != sdr.shape:
            raise ValueError(
                "User provided effective radius (`r_eff_um`) must have the same shape as `sdr`"
                f" {sdr.shape}"
            )

        r_eff_um_h = r_eff_um[idx0]

    # Local contrail shortwave radiative forcing calculations
    alpha_c = contrail_albedo(
        tau_contrail_h,
        mue,
        r_eff_um_h,
        A_mu,
        B_mu,
        C_mu,
        delta_sr,
        F_r,
        gamma_lower,
        gamma_upper,
    )

    e_sw = effective_tau_cirrus(tau_cirrus_h, mue, delta_sc, delta_sc_aps)

    # calculate the RF SW per habit type
    # see eqn (5) in :cite:`schumannParametricRadiativeForcing2012`
    rf_sw_per_habit = np.minimum(-sdr_h * ((t_a - albedo_) ** 2) * alpha_c * e_sw, 0.0)

    # Weight and sum the RF contributions of each habit type according the
    # habit weight regime at the waypoint
    # see eqn (12) in :cite:`schumannParametricRadiativeForcing2012`
    # use fancy indexing to re-assign values to 2d array of waypoint x habit type
    rf_sw_weighted = np.zeros_like(habit_weights_)
    rf_sw_weighted[idx0, idx1] = rf_sw_per_habit * habit_weights_[habit_weight_mask]

    return np.sum(rf_sw_weighted, axis=1)


def net_radiative_forcing(
    rf_lw: npt.NDArray[np.float_], rf_sw: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the local contrail net radiative forcing (rf_net).

    RF Net = Longwave RF (positive) + Shortwave RF (negative)

    Parameters
    ----------
    rf_lw : npt.NDArray[np.float_]
        local contrail longwave radiative forcing, [:math:`W m^{-2}`]
    rf_sw : npt.NDArray[np.float_]
        local contrail shortwave radiative forcing, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        local contrail net radiative forcing, [:math:`W m^{-2}`]
    """
    return rf_lw + rf_sw


def olr_reduction_natural_cirrus(
    tau_cirrus: npt.NDArray[np.float_], delta_lc: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate reduction in outgoing longwave radiation (OLR) due to the presence of natural cirrus.

    Natural cirrus has optical depth ``tau_cirrus`` above the contrail.
    See ``e_lw`` in Eq. (4) of Schumann et al. (2012).

    Parameters
    ----------
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the
        contrail for each waypoint.
    delta_lc : npt.NDArray[np.float_]
        Habit specific parameter to approximate the reduction of the outgoing
        longwave radiation at the contrail level due to natural cirrus above the contrail.

    Returns
    -------
    npt.NDArray[np.float_]
        Reduction of outgoing longwave radiation
    """
    # e_lw calculations
    return np.exp(-delta_lc * tau_cirrus)


def contrail_effective_emissivity(
    r_eff_um: npt.NDArray[np.float_], delta_lr: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    r"""Calculate the effective emissivity of the contrail, ``f_lw``.

    Refer to Eq. (3) of Schumann et al. (2012).

    Parameters
    ----------
    r_eff_um : npt.NDArray[np.float_]
        Effective radius for each waypoint, n_waypoints x 8 (habit) columns, [:math:`\mu m`]
        See :func:`effective_radius_habit`.
    delta_lr : npt.NDArray[np.float_]
        Habit specific parameter to approximate the effective emissivity of the contrail.

    Returns
    -------
    npt.NDArray[np.float_]
        Effective emissivity of the contrail
    """
    # f_lw calculations
    return 1.0 - np.exp(-delta_lr * r_eff_um)


def albedo(sdr: npt.NDArray[np.float_], rsr: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """
    Calculate albedo along contrail waypoint.

    Albedo, the diffuse reflection of solar radiation out of the total solar radiation,
    is computed based on the solar direct radiation (`sdr`) and reflected solar radiation (`rsr`).

    Output values range between 0 (corresponding to a black body that absorbs
    all incident radiation) and 1 (a body that reflects all incident radiation).

    Parameters
    ----------
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Albedo value, [:math:`[0 - 1]`]
    """
    day = sdr > 0.0
    albedo_ = np.zeros(sdr.shape)
    albedo_[day] = rsr[day] / sdr[day]
    albedo_.clip(0.0, 1.0, out=albedo_)
    return albedo_


def contrail_albedo(
    tau_contrail: npt.NDArray[np.float_],
    mue: npt.NDArray[np.float_],
    r_eff_um: npt.NDArray[np.float_],
    A_mu: npt.NDArray[np.float_],
    B_mu: npt.NDArray[np.float_],
    C_mu: npt.NDArray[np.float_],
    delta_sr: npt.NDArray[np.float_],
    F_r: npt.NDArray[np.float_],
    gamma_lower: npt.NDArray[np.float_],
    gamma_upper: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the contrail albedo, ``alpha_c``.

    Refer to Eq. (6) of Schumann et al. (2012),

    Parameters
    ----------
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    mue : npt.NDArray[np.float_]
        Cosine of the solar zenith angle (theta), mue = cos(theta) = sdr/sd0
    r_eff_um : npt.NDArray[np.float_]
        Effective radius for each waypoint, n_waypoints x 8 (habit) columns, [:math:`\mu m`]
        See :func:`effective_radius_habit`.
    A_mu : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the albedo of the contrail
    B_mu : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the SZA-dependent contrail sideward scattering
    C_mu : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the albedo of the contrail
    delta_sr : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the effective contrail optical depth
    F_r : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the effective contrail optical depth
    gamma_lower : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the contrail reflectances
    gamma_upper : npt.NDArray[np.float_]
        Habit-specific parameter to approximate the contrail reflectances

    Returns
    -------
    npt.NDArray[np.float_]
        Contrail albedo for each waypoint and ice particle habit
    """
    tau_aps = tau_contrail * (1.0 - F_r * (1 - np.exp(-delta_sr * r_eff_um)))
    tau_eff = tau_aps / (mue + 1e-6)
    r_c = 1.0 - np.exp(-gamma_upper * tau_eff)
    r_c_aps = np.exp(-gamma_lower * tau_eff)

    f_mu = (2.0 * (1.0 - mue)) ** B_mu - 1.0
    return r_c * (C_mu + (A_mu * r_c_aps * f_mu))


def effective_tau_cirrus(
    tau_cirrus: npt.NDArray[np.float_],
    mue: npt.NDArray[np.float_],
    delta_sc: npt.NDArray[np.float_],
    delta_sc_aps: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective optical depth of natural cirrus above the contrail, ``e_sw``.

    Refer to Eq. (11) of Schumann et al. (2012).

    Parameters
    ----------
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the
        contrail for each waypoint.
    mue : npt.NDArray[np.float_]
        Cosine of the solar zenith angle (theta), mue = cos(theta) = sdr/sd0
    delta_sc : npt.NDArray[np.float_]
        Habit-specific parameter to account for the optical depth of natural
        cirrus above the contrail
    delta_sc_aps : npt.NDArray[np.float_]
        Habit-specific parameter to account for the optical depth of natural
        cirrus above the contrail

    Returns
    -------
    npt.NDArray[np.float_]
        Effective optical depth of natural cirrus above the contrail,
        ``n_waypoints x 8`` (habit) columns.
    """
    tau_cirrus_eff = tau_cirrus / (mue + 1e-6)
    return np.exp(tau_cirrus * delta_sc) - (tau_cirrus_eff * delta_sc_aps)



"""Ensure gridded output can be written to disk."""

import pathlib

import numpy as np
import pytest
import xarray as xr

try:
    from pycontrails.models.cocipgrid import CocipGrid
except ImportError:
    pytest.skip("CocipGrid not available", allow_module_level=True)

from pycontrails import MetDataset
from pycontrails.models.humidity_scaling import ConstantHumidityScaling
from pycontrails.models.issr import ISSR
from pycontrails.models.pcr import PCR
from pycontrails.models.sac import SAC
from pycontrails.utils import temp
from tests import BADA3_PATH


@pytest.mark.skipif(not BADA3_PATH.is_dir(), reason="BADA3 not available")
@pytest.mark.parametrize("model", [SAC, ISSR, CocipGrid, PCR])
def test_model_output_to_netcdf(met_cocip1: MetDataset, rad_cocip1: MetDataset, model: type):
    """Ensure gridded output can be written to netCDF.

    This was a problem in 0.28.0 with humidity scaling metadata.
    """
    kwargs = {"met": met_cocip1, "humidity_scaling": ConstantHumidityScaling()}
    if issubclass(model, CocipGrid):
        kwargs["rad"] = rad_cocip1
        kwargs["max_age"] = np.timedelta64(4, "h")
        kwargs["interpolation_bounds_error"] = True
        kwargs["bada3_path"] = BADA3_PATH

    instance = model(**kwargs)

    coords = met_cocip1.coords

    # Do some slicing to keep everything in bounds
    coords["time"] = coords["time"][:1]
    coords["longitude"] = coords["longitude"][:-1]
    coords["latitude"] = coords["latitude"][1:-4]
    coords["level"] = [220, 230]

    source = MetDataset.from_coords(**coords)
    out = instance.eval(source)

    with temp.temp_file() as temp_file:
        path = pathlib.Path(temp_file)

        da = out.data
        if isinstance(instance, CocipGrid):
            da = da["ef_per_m"]
            da.attrs.update(out.attrs)

        assert np.all(np.isfinite(da))

        assert "pycontrails_version" in da.attrs
        assert "humidity_scaling_name" in da.attrs

        assert isinstance(da, xr.DataArray)
        assert isinstance(da, xr.DataArray)
        assert not path.is_file()

        da.to_netcdf(temp_file)
        assert path.is_file()
        assert path.stat().st_size > 10000

        assert da.attrs == xr.open_dataarray(temp_file).attrs

    assert not path.is_file()



from __future__ import annotations

import os
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import DiskCacheStore, MetDataset, MetVariable
from pycontrails.core.met_var import AirTemperature, SurfacePressure
from pycontrails.datalib.ecmwf import ECMWF_VARIABLES, ERA5, HRES
from pycontrails.datalib.ecmwf.hres import get_forecast_filename


def test_environ_keys() -> None:
    """Test CDS Keys."""
    assert os.environ.get("CDSAPI_URL") == "FAKE"
    assert os.environ.get("CDSAPI_KEY") == "FAKE"


def test_ECMWF_variables() -> None:
    """Test ECMWF_VARIABLES."""
    assert isinstance(ECMWF_VARIABLES[0], MetVariable)
    assert isinstance(AirTemperature, MetVariable)

    # confirm properties
    assert AirTemperature.ecmwf_id == 130
    assert AirTemperature.standard_name == "air_temperature"
    assert AirTemperature.short_name == "t"
    assert AirTemperature.units == "K"
    assert AirTemperature.ecmwf_link == "https://apps.ecmwf.int/codes/grib/param-db?id=130"


def test_ERA5_time_input() -> None:
    """Test TimeInput parsing."""
    # accept single time
    era5 = ERA5(time=datetime(2019, 5, 31, 0), variables=["vo"], pressure_levels=[200])
    assert era5.timesteps == [datetime(2019, 5, 31, 0)]
    era5 = ERA5(time=[datetime(2019, 5, 31, 0)], variables=["vo"], pressure_levels=[200])
    assert era5.timesteps == [datetime(2019, 5, 31, 0)]

    # accept single time with minutes defined
    era5 = ERA5(time=datetime(2019, 5, 31, 0, 29), variables=["vo"], pressure_levels=[200])
    assert era5.timesteps == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]
    era5 = ERA5(time=[datetime(2019, 5, 31, 0, 29)], variables=["vo"], pressure_levels=[200])
    assert era5.timesteps == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]

    # throw ValueError for length == 0
    with pytest.raises(ValueError):
        ERA5(time=[], variables=["vo"], pressure_levels=[200])

    # throw ValueError for length > 2
    with pytest.raises(ValueError):
        ERA5(
            time=[datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 0)],
            variables=["vo"],
            pressure_levels=[200],
        )

    # accept (start, end)
    era5 = ERA5(
        time=(datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 3)),
        variables=["vo"],
        pressure_levels=[200],
    )
    assert era5.timesteps == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]
    era5 = ERA5(
        time=(datetime(2019, 5, 31, 0, 29), datetime(2019, 5, 31, 2, 40)),
        variables=["vo"],
        pressure_levels=[200],
    )
    assert era5.timesteps == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]

    # support alternate types for input
    era5 = ERA5(
        time=pd.to_datetime(datetime(2019, 5, 31, 0, 29)),
        variables=["vo"],
        pressure_levels=[200],
    )
    assert era5.timesteps == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]
    era5 = ERA5(time=np.datetime64("2019-05-31T00:29:00"), variables=["vo"], pressure_levels=[200])
    assert era5.timesteps == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]
    era5 = ERA5(
        time=(np.datetime64("2019-05-31T00:29:00"), np.datetime64("2019-05-31T02:40:00")),
        variables=["vo"],
        pressure_levels=[200],
    )
    assert era5.timesteps == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]


def test_ERA5_inputs() -> None:
    """Test ERA5 __init__."""
    with pytest.raises(ValueError):  # bad pressure levels
        ERA5(time=[], variables=[], pressure_levels=[17, 18, 19])

    e1 = ERA5(time=datetime(2000, 1, 1), variables="vo", pressure_levels=200)
    e2 = ERA5(time=[datetime(2000, 1, 1)], variables=["vo"], pressure_levels=[200])
    e3 = ERA5(
        time=np.datetime64("2000-01-01 00:00:00"),
        variables=["vo"],
        pressure_levels=[200],
    )
    e4 = ERA5(
        time=np.array([np.datetime64("2000-01-01 00:00:00")]),
        variables=["vo"],
        pressure_levels=[200],
    )
    assert e1.variables == e2.variables
    assert e1.pressure_levels == e2.pressure_levels
    assert e1.timesteps == e2.timesteps == e3.timesteps == e4.timesteps


def test_ERA5_repr() -> None:
    """Test ERA5 __repr__."""
    era5 = ERA5(time=datetime(2000, 1, 1), variables="vo", pressure_levels=200)
    out = repr(era5)
    assert "ERA5" in out
    assert "Dataset:" in out


def test_ERA5_cachestore(met_ecmwf_pl_path: str, override_cache: DiskCacheStore) -> None:
    """Test ERA5 cachestore input."""

    # clear cache to start
    override_cache.clear()

    era5 = ERA5(
        time=datetime(2000, 1, 1),
        variables="vo",
        pressure_levels=200,
        cachestore=override_cache,
    )
    cachepath = era5.create_cachepath(datetime(2000, 1, 1))
    assert "20000101-00-era5pl0.25reanalysis.nc" in cachepath

    # load actual data
    assert override_cache.size == 0
    times = [datetime(2019, 5, 31, 5, 0, 0)]
    pressure_levels = [300, 250]
    variables = ["air_temperature", "relative_humidity", "specific_humidity"]
    era5 = ERA5(
        time=times,
        variables=variables,
        pressure_levels=pressure_levels,
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    mds = era5.open_metdataset()
    assert isinstance(mds, MetDataset)
    assert override_cache.size > 0

    # allow cache to be None
    pre_init_size = DiskCacheStore().size
    era5 = ERA5(
        time=times,
        variables=variables,
        pressure_levels=pressure_levels,
        paths=met_ecmwf_pl_path,
        cachestore=None,
    )
    mds = era5.open_metdataset()
    assert isinstance(mds, MetDataset)
    post_init_size = DiskCacheStore().size
    assert pre_init_size == post_init_size


def test_ERA5_pressure_levels(met_ecmwf_pl_path: str, override_cache: DiskCacheStore) -> None:
    """Test ERA5 pressure_level parsing."""
    times = [datetime(2019, 5, 31, 5, 0, 0)]
    pressure_levels = [300, 250]
    variables = ["air_temperature", "relative_humidity", "specific_humidity"]
    era5 = ERA5(time=times, variables=variables, pressure_levels=pressure_levels)

    # properties
    assert 300 in era5.pressure_levels
    assert AirTemperature in era5.pressure_level_variables
    assert AirTemperature in era5.supported_variables
    assert isinstance(era5._cachepaths[0], str) and "2019" in era5._cachepaths[0]

    # load
    era5 = ERA5(
        time=times,
        variables=variables,
        pressure_levels=pressure_levels,
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    met = era5.open_metdataset()
    assert isinstance(met, MetDataset)

    # preprocess
    assert met.data.attrs["pycontrails_version"] == "0.34.0"
    assert met.data["relative_humidity"].attrs["long_name"] == "Relative humidity"
    assert met.data["relative_humidity"].attrs.get("_pycontrails_modified", False)
    assert met.data["relative_humidity"].attrs["units"] == "[0 - 1]"
    assert 250 in met.data["level"] and 200 not in met.data["level"]
    assert met.data.longitude.min().values == -160

    # consistency
    assert np.all(met.data["level"].values == sorted(pressure_levels))
    # dealing with discrepancy between datetime and np.datetime64
    assert np.all(met.data["time"].values == np.array([np.datetime64(t) for t in era5.timesteps]))


def test_ERA5_single_levels(met_ecmwf_sl_path: str, override_cache: DiskCacheStore) -> None:
    """Test ERA5 surface level parsing."""
    times = [datetime(2019, 5, 31, 5, 0, 0)]
    variables = ["surface_air_pressure"]
    era5 = ERA5(times, variables)

    # init
    assert era5.dataset == "reanalysis-era5-single-levels"
    assert era5.pressure_levels == [-1]

    # properties
    assert SurfacePressure in era5.single_level_variables
    assert SurfacePressure in era5.supported_variables
    assert isinstance(era5._cachepaths[0], str) and "2019" in era5._cachepaths[0]

    # load
    era5 = ERA5(time=times, variables=variables, paths=met_ecmwf_sl_path, cachestore=override_cache)
    met = era5.open_metdataset()
    assert isinstance(met, MetDataset)

    # preprocess
    assert met.data.attrs["pycontrails_version"] == "0.34.0"
    assert "surface_air_pressure" in met.data
    assert -1 in met.data["level"] and 200 not in met.data["level"]
    assert met.data.longitude.min().values == -160

    # consistency
    assert np.all(met.data["level"].values == era5.pressure_levels)
    # dealing with discrepancy between datetime and np.datetime64
    assert np.all(met.data["time"].values == np.array([np.datetime64(t) for t in era5.timesteps]))


def test_ERA5_radiation_processing(rad_cocip1: MetDataset) -> None:
    """Test ERA5 radiation processing, as displaying in cocip rad data."""

    assert rad_cocip1.data["top_net_thermal_radiation"].attrs.get("_pycontrails_modified", False)
    assert rad_cocip1.data["top_net_solar_radiation"].attrs.get("_pycontrails_modified", False)

    assert rad_cocip1.data["top_net_thermal_radiation"].attrs["units"] == "W m**-2"
    assert rad_cocip1.data["top_net_solar_radiation"].attrs["units"] == "W m**-2"


def test_ERA5_hash() -> None:
    """Test ERA5 hash."""

    era5 = ERA5(time=datetime(2022, 3, 5), variables=["t", "r"], pressure_levels=200)
    era52 = ERA5(time=datetime(2022, 3, 5), variables=["t", "r"], pressure_levels=[200, 300])
    assert era5.hash == "6939450efa6c348471dea359a0055f49d72922e6"
    assert era52.hash == "58af5be8d2826bbacee14aa471da6d51060b4f8c"
    assert era5.hash != era52.hash


def test_ERA5_paths_with_time(
    met_ecmwf_pl_path: str,
    met_ecmwf_sl_path: str,
    override_cache: DiskCacheStore,
) -> None:
    """Test ERA5 paths input with time."""

    time = datetime(2019, 5, 31, 5, 0, 0)

    # these should both work without warning
    era5pl = ERA5(
        time=time,
        variables=["air_temperature"],
        pressure_levels=[300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    metpl = era5pl.open_metdataset()
    assert metpl.data.attrs["pycontrails_version"] == "0.34.0"

    era5sl = ERA5(
        time=time,
        variables=["surface_air_pressure"],
        paths=met_ecmwf_sl_path,
        cachestore=override_cache,
    )
    metsl = era5sl.open_metdataset()
    assert metsl.data.attrs["pycontrails_version"] == "0.34.0"

    # these files should be getting cached by default
    assert override_cache.size > 0


def test_ERA5_paths_without_time(
    met_ecmwf_pl_path: str,
    override_cache: DiskCacheStore,
) -> None:
    """Test ERA5 paths input with time=None."""

    era5pl = ERA5(
        time=None,
        variables=["air_temperature"],
        pressure_levels=[300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    metpl = era5pl.open_metdataset()
    ds = xr.open_dataset(met_ecmwf_pl_path)  # open manually for comparison
    assert era5pl.timesteps and len(era5pl.timesteps) == len(ds["time"])
    assert (metpl.data["time"].values == ds["time"].values).all()

    # allow cache to be None
    pre_init_size = DiskCacheStore().size
    era5pl = ERA5(
        time=None,
        variables=["air_temperature"],
        pressure_levels=[300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=None,
    )
    metpl = era5pl.open_metdataset()
    post_init_size = DiskCacheStore().size
    assert pre_init_size == post_init_size


def test_ERA5_paths_with_error(
    met_ecmwf_pl_path: str,
    override_cache: DiskCacheStore,
) -> None:
    """Test ERA5 issues errors with paths input."""

    era5pl = ERA5(
        time="2019-06-02 00:00:00, ",
        variables=["air_temperature"],
        pressure_levels=[300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    with pytest.raises(KeyError, match="2019-06-02T00:00:00"):
        era5pl.open_metdataset()

    time = datetime(2019, 5, 31, 5, 0, 0)

    # check variables
    era5pl = ERA5(
        time=time,
        variables=["z"],
        pressure_levels=[300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    with pytest.raises(KeyError, match="z"):
        era5pl.open_metdataset()

    # check pressure levels
    era5pl = ERA5(
        time=time,
        variables=["air_temperature"],
        pressure_levels=[400, 300, 250],
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    with pytest.raises(KeyError, match="400"):
        era5pl.open_metdataset()


def test_ERA5_dataset(met_ecmwf_pl_path: str, met_ecmwf_sl_path: str) -> None:
    """Test ERA5 paths input."""

    dspl = xr.open_dataset(met_ecmwf_pl_path)
    times = datetime(2019, 5, 31, 5, 0, 0)

    # these should both work without warning
    era5pl = ERA5(
        time=times,
        variables=["air_temperature"],
        pressure_levels=[300, 250],
    )
    metpl = era5pl.open_metdataset(dataset=dspl)

    assert metpl.data.attrs["pycontrails_version"] == "0.34.0"
    assert 250 in metpl.data["level"].values and 300 in metpl.data["level"].values

    dssl = xr.open_dataset(met_ecmwf_sl_path)
    era5sl = ERA5(time=times, variables=["surface_air_pressure"])
    metsl = era5sl.open_metdataset(dataset=dssl)
    assert metsl.data.attrs["pycontrails_version"] == "0.34.0"
    assert -1 in metsl.data["level"].values
    assert "surface_air_pressure" in metsl.data.data_vars

    # make sure there is sanity checking

    # check time
    era5pl = ERA5(
        time="2019-06-02 00:00:00, ",
        variables=["air_temperature"],
        pressure_levels=[300, 250],
    )
    with pytest.raises(KeyError, match="2019-06-02T00:00:00"):
        metpl = era5pl.open_metdataset(dataset=dspl)

    # check variables
    era5pl = ERA5(
        time=times,
        variables=["z"],
        pressure_levels=[300, 250],
    )
    with pytest.raises(KeyError, match="z"):
        metpl = era5pl.open_metdataset(dataset=dspl)

    # check pressure levels
    era5pl = ERA5(
        time=times,
        variables=["air_temperature"],
        pressure_levels=[400, 300, 250],
    )
    with pytest.raises(KeyError, match="400"):
        metpl = era5pl.open_metdataset(dataset=dspl)


def test_HRES_repr() -> None:
    """Test HRES __repr__."""
    hres = HRES(time=datetime(2000, 1, 1), variables="vo", pressure_levels=200)
    out = repr(hres)
    assert "HRES" in out
    assert "Forecast time:" in out


def test_HRES_inputs() -> None:
    """Test HRES time parsing."""
    with pytest.raises(ValueError):  # bad pressure levels
        HRES(time=[], variables=[], pressure_levels=[17, 18, 19])

    times = (datetime(2019, 5, 31, 2, 29), datetime(2019, 5, 31, 4, 29))
    pressure_levels = [300, 250]
    variables = ["air_temperature", "relative_humidity", "specific_humidity"]
    hres = HRES(times, variables, pressure_levels=pressure_levels)

    assert hres.forecast_time == datetime(2019, 5, 31, 0, 0)
    assert hres.step_offset == 2
    assert hres.steps == [2, 3, 4, 5]

    times_str = ("2019-05-31 02:29:00", "2019-05-31 04:29:00")
    pressure_levels = [300, 250]
    variables = ["air_temperature", "relative_humidity", "specific_humidity"]
    forecast_time = "2019-05-30 12:00:00"
    hres = HRES(
        time=times_str,
        variables=variables,
        pressure_levels=pressure_levels,
        forecast_time=forecast_time,
    )

    assert hres.forecast_time == datetime(2019, 5, 30, 12, 0)
    assert hres.step_offset == 14
    assert hres.steps == [14, 15, 16, 17]


def test_HRES_hash() -> None:
    """Test HRES hash."""

    hres = HRES(time=datetime(2022, 3, 5), variables=["t", "r"], pressure_levels=200)
    hres2 = HRES(time=datetime(2022, 3, 5), variables=["t", "r"], pressure_levels=[200, 300])
    assert hres.hash == "9f6c8ef71eed695d9e79093ee4cf750dec00895a"
    assert hres2.hash == "178f8cb671e8ac9e5ba086db7cc888d467ec11c2"
    assert hres.hash != hres2.hash


def test_HRES_dissemination_filename() -> None:
    """Test HRES dissemination filename creation."""

    forecast_time = datetime(2022, 12, 1, 0)

    timestep = datetime(2022, 12, 2, 8)
    forecast_time_str = get_forecast_filename(forecast_time, timestep)
    assert forecast_time_str == "A1D12010000120208001"

    # 0th step
    timestep = datetime(2022, 12, 1, 0)
    forecast_time_str = get_forecast_filename(forecast_time, timestep)
    assert forecast_time_str == "A1D12010000120100011"

    # 6th hour forecast
    forecast_time = datetime(2022, 12, 1, 6)
    timestep = datetime(2022, 12, 2, 0)
    forecast_time_str = get_forecast_filename(forecast_time, timestep)
    assert forecast_time_str == "A1S12010600120200001"

    # 6th hour forecast, 0th step
    forecast_time = datetime(2022, 12, 1, 6)
    timestep = datetime(2022, 12, 1, 6)
    forecast_time_str = get_forecast_filename(forecast_time, timestep)
    assert forecast_time_str == "A1S12010600120106011"

    # forecast time error
    with pytest.raises(ValueError, match="hour 0, 6, 12"):
        forecast_time = datetime(2022, 12, 1, 2)
        timestep = datetime(2022, 12, 1, 6)
        forecast_time_str = get_forecast_filename(forecast_time, timestep)

    # timestep before forecast
    with pytest.raises(ValueError, match="timestep must be on or after forecast time"):
        forecast_time = datetime(2022, 12, 2, 0)
        timestep = datetime(2022, 12, 1, 6)
        forecast_time_str = get_forecast_filename(forecast_time, timestep)



"""Convienence types."""

from __future__ import annotations

import functools
from datetime import datetime
from typing import Any, Callable, Type, TypeVar, Union

import numpy as np
import pandas as pd
import xarray as xr

#: Array like (np.ndarray, xr.DataArray)
ArrayLike = TypeVar("ArrayLike", np.ndarray, xr.DataArray, Union[xr.DataArray, np.ndarray])

#: Array or Float (np.ndarray, float)
ArrayOrFloat = TypeVar("ArrayOrFloat", np.ndarray, float)

#: Array like input (np.ndarray, xr.DataArray, np.float64, float)
ArrayScalarLike = TypeVar(
    "ArrayScalarLike",
    np.ndarray,
    xr.DataArray,
    np.float64,
    float,
    Union[np.ndarray, float],
    Union[xr.DataArray, np.ndarray],
)

#: Datetime like input (datetime, pd.Timestamp, np.datetime64)
DatetimeLike = TypeVar("DatetimeLike", datetime, pd.Timestamp, np.datetime64)


def support_arraylike(
    func: Callable[[np.ndarray], np.ndarray]
) -> Callable[[ArrayScalarLike], ArrayScalarLike]:
    """Extend a numpy universal function operating on arrays of floats.

    This decorator allows `func` to support any ArrayScalarLike parameter and
    keeps the return type consistent with the parameter.

    Parameters
    ----------
    func : Callable[[ArrayScalarLike], np.ndarray]
        A numpy `ufunc` taking in a single array with `float`-like dtype.
        This decorator assumes `func` returns a numpy array.

    Returns
    -------
    Callable[[ArrayScalarLike], ArrayScalarLike]
        Extended function.

    See Also
    --------
    - `numpy ufuncs <https://numpy.org/doc/stable/reference/ufuncs.html>`_
    """

    @functools.wraps(func)
    def wrapped(arr: ArrayScalarLike) -> ArrayScalarLike:
        x = np.asarray(arr)

        # Convert to float if not already
        if x.dtype not in (np.float32, np.float64):
            x = x.astype(np.float64)
        ret = func(x)

        # Numpy in, numpy out
        if isinstance(arr, np.ndarray):
            return ret

        # Keep python native numeric types native
        if isinstance(arr, (float, int, np.float64)):
            return ret.item()

        # Recreate pd.Series
        if isinstance(arr, pd.Series):
            return pd.Series(data=ret, index=arr.index)

        # Recreate xr.DataArray
        if isinstance(arr, xr.DataArray):
            return arr.copy(data=ret)  # See documentation for xr.copy!

        # Pass numpy `ret` through for anything else
        return ret

    return wrapped


def apply_nan_mask_to_arraylike(arr: ArrayLike, nan_mask: np.ndarray) -> ArrayLike:
    """Apply ``nan_mask`` to ``arr`` while maintaining the type.

    The parameter ``arr`` should have a ``float`` ``dtype``.

    This function is tested against :class:`xr.DataArray`, :class:`pd.Series`, and
    :class:`np.ndarray` types.

    Parameters
    ----------
    arr : ArrayLike
        Array with ``np.float64`` entries
    nan_mask : np.ndarray
        Boolean array of the same shape as ``arr``

    Returns
    -------
    ArrayLike
        Array ``arr`` with values in ``nan_mask`` set to ``np.nan``. The ``arr`` is
        mutated in place if it is a :class:`np.ndarray`. For :class:`xr.DataArray`,
        a copy is returned.

    Notes
    -----
    When ``arr`` is a :class:`xr.DataArray`, this function keeps any ``attrs``
    from ``arr`` in the returned instance.
    """
    if isinstance(arr, xr.DataArray):
        # The previous implementation uses xr.where instead of arr.where
        # There was some change in xarray 2022.6.0 that broke the former implementation
        # Instead, use arr.where
        return arr.where(~nan_mask, np.nan)

    # If we want to avoid copying, use np.where(~nan_mask, arr, np.nan)
    arr[nan_mask] = np.nan
    return arr


_Object = TypeVar("_Object")


def type_guard(
    obj: Any,
    type_: Type[_Object] | tuple[Type[_Object], ...],
    error_message: str | None = None,
) -> _Object:
    """Shortcut utility to type guard a variable with custom error message.

    Parameters
    ----------
    obj : Any
        Any variable object
    type_ : Type[_Object]
        Type of variable.
        Can be a tuple of types
    error_message : str, optional
        Custom error message

    Returns
    -------
    _Object
        Returns the input object ensured to be ``type_``

    Raises
    ------
    ValueError
        Raises ValueError if ``obj`` is not ``type_``
    """
    if not isinstance(obj, type_):
        raise ValueError(error_message or f"Object must be of type {type_}")

    return obj



"""Test `cocip_uncertainty` module."""

from __future__ import annotations

import numpy as np
import pytest
from scipy.stats.distributions import rv_frozen

from pycontrails.models.cocip import CocipParams
from pycontrails.models.cocip.cocip_uncertainty import CocipUncertaintyParams, habit_dirichlet

seeds = [None, 1, 2, 3, 5, 8]


def _get_overriden(c: CocipUncertaintyParams) -> list[str]:
    """Get overriden parameters."""
    # return [split[0] for param in c.as_dict() if len(split := param.split("_uncertainty")) == 2]
    ret = []
    for param in c.as_dict():
        split = param.split("_uncertainty")
        if len(split) == 2:
            ret.append(split[0])
    return ret


@pytest.mark.parametrize("seed", seeds)
def test_cocip_uncertainty_overrides(seed: int | None) -> None:
    """Check that `CocipUncertaintyParams` overrides values from `CocipParams`."""
    c1 = CocipParams()
    c2 = CocipUncertaintyParams(seed=seed)
    for param in _get_overriden(c2):
        assert getattr(c1, param) != getattr(c2, param)


@pytest.mark.parametrize("seed", seeds)
def test_cocip_uncertainty_ranges(seed: int | None) -> None:
    """Check that parameter values are contained in specified ranges."""
    c = CocipUncertaintyParams(seed=seed)
    for param in _get_overriden(c):
        value = getattr(c, param)
        uncertainty_param = getattr(c, f"{param}_uncertainty")
        if uncertainty_param.dist.name in ["uniform", "triang"]:
            assert value > uncertainty_param.kwds["loc"]
            assert value < uncertainty_param.kwds["loc"] + uncertainty_param.kwds["scale"]
        elif uncertainty_param.dist.name == "halfnorm":
            assert value > uncertainty_param.kwds["loc"]
        else:
            assert uncertainty_param.dist.name in ["norm", "lognorm"]


def test_uncertainty_seed() -> None:
    """Check that random value generation is reproducible via the seed parameter."""
    c1 = CocipUncertaintyParams(seed=42)
    c2 = CocipUncertaintyParams(seed=42)
    for param in _get_overriden(c1):
        assert getattr(c1, param) == getattr(c2, param)

    c3 = CocipUncertaintyParams()
    for param in _get_overriden(c1):
        assert getattr(c1, param) != getattr(c3, param)


def test_unspecified_seed() -> None:
    """Check that the default value of seed (None) gives random behavior."""
    c1 = CocipUncertaintyParams()
    c2 = CocipUncertaintyParams()
    for param in _get_overriden(c1):
        assert getattr(c1, param) != getattr(c2, param)


def test_rng_class_variable() -> None:
    """Test class variable `rng`."""
    c1 = CocipUncertaintyParams(seed=5)
    state1 = CocipUncertaintyParams.rng.__getstate__()
    c2 = CocipUncertaintyParams()
    state2 = CocipUncertaintyParams.rng.__getstate__()
    assert state1 != state2

    c3 = CocipUncertaintyParams(seed=5)
    assert state1 == CocipUncertaintyParams.rng.__getstate__()
    c4 = CocipUncertaintyParams()
    assert state2 == CocipUncertaintyParams.rng.__getstate__()

    for param in _get_overriden(c1):
        assert getattr(c1, param) == getattr(c3, param)
        assert getattr(c2, param) == getattr(c4, param)


def test_mean_agrees_with_certain_defaults() -> None:
    """Check that the uncertainty distribution mean agrees with the default `CocipParams` value."""
    c1 = CocipParams()
    c2 = CocipUncertaintyParams()
    for param, dist in c2.uncertainty_params.items():
        if param in [
            "wind_shear_enhancement_exponent",
            "initial_wake_vortex_depth",
            "sedimentation_impact_factor",
            "nvpm_ei_n_enhancement_factor",
            "rf_sw_enhancement_factor",
            "rf_lw_enhancement_factor",
        ]:
            assert dist.mean() == getattr(c1, param)
        elif param == "rhi_adj":
            assert dist.mean() == 0.97
        else:
            assert param in ["rhi_boost_exponent", "habit_distributions"]


def test_uncertainty_default_dist_within_expect_cocip_function_domain() -> None:
    """Check that default distribution fall within the domains of `Cocip` methods."""
    c = CocipUncertaintyParams(seed=1001)
    n = 1_000_000

    # initial wake vortex depth should be nonnegative
    assert np.all(c.initial_wake_vortex_depth_uncertainty.rvs(n) >= 0)

    # nvpm_ei_n_enhancement_factor should be between 0.3 and 3
    assert np.all(c.nvpm_ei_n_enhancement_factor_uncertainty.rvs(n) >= 0.3)
    assert np.all(c.nvpm_ei_n_enhancement_factor_uncertainty.rvs(n) <= 3)

    # habit_distributions should all sum to 1 and have the right shape
    assert np.all(np.round(np.sum(c.habit_distributions_uncertainty.rvs(), axis=1), 3) == 1)

    # rf_lw_enhancement_factor should be bounded near 1
    assert np.all(c.rf_lw_enhancement_factor_uncertainty.rvs(n) > 0.5)
    assert np.all(c.rf_lw_enhancement_factor_uncertainty.rvs(n) < 1.5)

    # rf_sw_enhancement_factor should be bounded near 1
    assert np.all(c.rf_sw_enhancement_factor_uncertainty.rvs(n) > 0.4)
    assert np.all(c.rf_sw_enhancement_factor_uncertainty.rvs(n) < 1.6)


def test_habit_dirichlet() -> None:
    """Test custom habit dirichlet distribution generator."""

    hd = habit_dirichlet()

    # fakes an instance of rv_frozen
    assert isinstance(hd, rv_frozen)

    # defaults to C == 96
    assert hd.C == 96

    # has a rvs() method
    habit_distributions = hd.rvs()
    assert np.all(np.round(np.sum(habit_distributions, axis=1), 4) == 1)

    # doesn't support n
    with pytest.raises(ValueError, match="only supports creating one rv"):
        hd.rvs(4)

    # only support random_state
    with pytest.raises(ValueError, match="only supports creating one rv"):
        habit_distributions = hd.rvs(size=2)

    habit_distributions = hd.rvs(size=1)
    assert habit_distributions.shape == (6, 8)

    # accessible via habit_distribution_uncertainty
    c = CocipUncertaintyParams(seed=1001)
    assert isinstance(c.habit_distributions_uncertainty, rv_frozen)
    habit_distributions = c.habit_distributions_uncertainty.rvs()
    assert np.all(np.round(np.sum(habit_distributions, axis=1), 4) == 1)

    cp = CocipParams()
    radius_threshold_um = cp.radius_threshold_um
    habits = cp.habits
    assert habit_distributions.shape == (radius_threshold_um.size + 1, habits.size)


def test_none_distribution() -> None:
    """Test when distribution is set to None."""

    params = CocipUncertaintyParams().as_dict()
    assert params["rf_lw_enhancement_factor"] != 1
    assert params["rf_sw_enhancement_factor"] != 1

    params = CocipUncertaintyParams(
        rf_lw_enhancement_factor_uncertainty=None, rf_sw_enhancement_factor_uncertainty=None
    ).as_dict()
    assert params["rf_lw_enhancement_factor"] == 1
    assert params["rf_sw_enhancement_factor"] == 1



"""Temp utilities."""

import logging
import os
import tempfile
from contextlib import contextmanager
from typing import Generator

logger = logging.getLogger(__name__)


def temp_filename() -> str:
    """Get a filename in the host computers temp directory.

    More robust than using tempfile.NamedTemporaryFile()

    Returns
    -------
    str
        Temp filename

    See Also
    --------
    temp_file : Context manager for temp file creation and cleanup
    """
    return os.path.join(tempfile.gettempdir(), os.urandom(24).hex())


def remove_tempfile(temp_filename: str) -> None:
    """Remove temp file.

    Parameters
    ----------
    temp_filename : str
        Temp filename
    """
    try:
        os.unlink(temp_filename)
    except OSError as e:
        logger.warning("Failed to delete temp file %s with error %s", temp_filename, e)


@contextmanager
def temp_file() -> Generator[str, None, None]:
    """Get context manager for temp file creation and cleanup."""
    filename = temp_filename()
    try:
        yield filename
    finally:
        remove_tempfile(filename)
